# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.providers.common.compat.sdk import AirflowSkipException
from airflow.providers.slack.transfers.sql_to_slack import SqlToSlackApiFileOperator

from tests_common.test_utils.compat import timezone

TEST_DAG_ID = "sql_to_slack_unit_test"
TEST_TASK_ID = "sql_to_slack_unit_test_task"
DEFAULT_DATE = timezone.datetime(2017, 1, 1)


class TestSqlToSlackApiFileOperator:
    def setup_method(self):
        self.default_op_kwargs = {
            "sql": "SELECT 1",
            "sql_conn_id": "test-sql-conn-id",
            "slack_conn_id": "test-slack-conn-id",
            "sql_hook_params": None,
            "parameters": None,
        }

    @mock.patch("airflow.providers.slack.transfers.sql_to_slack.BaseSqlToSlackOperator._get_query_results")
    @mock.patch("airflow.providers.slack.transfers.sql_to_slack.SlackHook")
    @pytest.mark.parametrize(
        ("filename", "df_method"),
        [
            ("awesome.json", "to_json"),
            ("awesome.json.zip", "to_json"),
            ("awesome.csv", "to_csv"),
            ("awesome.csv.xz", "to_csv"),
            ("awesome.html", "to_html"),
        ],
    )
    @pytest.mark.parametrize("df_kwargs", [None, {}, {"foo": "bar"}])
    @pytest.mark.parametrize("channels", ["#random", "#random,#general", None])
    @pytest.mark.parametrize("initial_comment", [None, "Test Comment"])
    @pytest.mark.parametrize("title", [None, "Test File Title"])
    @pytest.mark.parametrize(
        ("slack_op_kwargs", "hook_extra_kwargs"),
        [
            pytest.param(
                {},
                {"base_url": None, "timeout": None, "proxy": None, "retry_handlers": None},
                id="default-hook-parameters",
            ),
            pytest.param(
                {
                    "slack_base_url": "https://foo.bar",
                    "slack_timeout": 42,
                    "slack_proxy": "http://spam.egg",
                    "slack_retry_handlers": [],
                },
                {
                    "base_url": "https://foo.bar",
                    "timeout": 42,
                    "proxy": "http://spam.egg",
                    "retry_handlers": [],
                },
                id="with-extra-hook-parameters",
            ),
        ],
    )
    def test_send_file(
        self,
        mock_slack_hook_cls,
        mock_get_query_results,
        filename,
        df_method,
        df_kwargs,
        channels,
        initial_comment,
        title,
        slack_op_kwargs: dict,
        hook_extra_kwargs: dict,
    ):
        # Mock Hook
        mock_send_file = mock.MagicMock()
        setattr(mock_slack_hook_cls.return_value, "send_file_v1_to_v2", mock_send_file)

        # Mock returns pandas.DataFrame and expected method
        mock_df = mock.MagicMock()
        mock_df_output_method = mock.MagicMock()
        setattr(mock_df, df_method, mock_df_output_method)
        mock_get_query_results.return_value = mock_df

        op_kwargs = {
            **self.default_op_kwargs,
            "slack_conn_id": "expected-test-slack-conn-id",
            "slack_filename": filename,
            "slack_channels": channels,
            "slack_initial_comment": initial_comment,
            "slack_title": title,
            "df_kwargs": df_kwargs,
            **slack_op_kwargs,
        }
        op = SqlToSlackApiFileOperator(task_id="test_send_file", **op_kwargs)

        mock.patch("airflow.providers.slack.transfers.sql_to_slack.SlackHook")
        op.execute(mock.MagicMock())

        mock_slack_hook_cls.assert_called_once_with(
            slack_conn_id="expected-test-slack-conn-id", **hook_extra_kwargs
        )
        mock_get_query_results.assert_called_once_with()
        mock_df_output_method.assert_called_once_with(mock.ANY, **(df_kwargs or {}))
        mock_send_file.assert_called_once_with(
            channels=channels,
            filename=filename,
            initial_comment=initial_comment,
            title=title,
            file=mock.ANY,
        )

    @pytest.mark.parametrize(
        "filename",
        [
            "foo.parquet",
            "bat.parquet.snappy",
            "spam.xml",
            "egg.xlsx",
        ],
    )
    def test_unsupported_format(self, filename):
        op = SqlToSlackApiFileOperator(
            task_id="test_send_file", slack_filename=filename, **self.default_op_kwargs
        )
        with pytest.raises(ValueError, match="Unsupported file format"):
            op.execute(mock.MagicMock())

    @mock.patch("airflow.providers.slack.transfers.sql_to_slack.SlackHook")
    @mock.patch("airflow.providers.slack.transfers.sql_to_slack.BaseSqlToSlackOperator._get_query_results")
    def test_null_output_sending_empty_file_by_default(self, mock_get_query_results, mock_slack_hook_cls):
        op_kwargs = {
            **self.default_op_kwargs,
            "slack_conn_id": "expected-test-slack-conn-id",
            "slack_filename": "test_filename.csv",
            "slack_channels": ["#random"],
            "slack_initial_comment": "test_comment",
            "slack_title": "test_title",
        }
        op = SqlToSlackApiFileOperator(task_id="test_send_file", **op_kwargs)

        # Mock empty query results
        mock_df = mock.MagicMock()
        mock_df.configure_mock(**{"empty.return_value": True})
        mock_get_query_results.return_value = mock_df

        op.execute(mock.MagicMock)
        mock_slack_hook_cls.assert_called_once()

    @mock.patch("airflow.providers.slack.transfers.sql_to_slack.SlackHook")
    @mock.patch("airflow.providers.slack.transfers.sql_to_slack.BaseSqlToSlackOperator._get_query_results")
    def test_null_output_skip_sending_file(self, mock_get_query_results, mock_slack_hook_cls):
        op_kwargs = {
            **self.default_op_kwargs,
            "slack_conn_id": "expected-test-slack-conn-id",
            "slack_filename": "test_filename.csv",
            "slack_channels": ["#random"],
            "slack_initial_comment": "test_comment",
            "slack_title": "test_title",
            "action_on_empty_df": "skip",
        }
        op = SqlToSlackApiFileOperator(task_id="test_send_file", **op_kwargs)

        # Mock empty query results
        mock_df = mock.MagicMock()
        mock_df.configure_mock(**{"empty.return_value": True})
        mock_get_query_results.return_value = mock_df

        with pytest.raises(AirflowSkipException):
            op.execute(mock.MagicMock())
        mock_slack_hook_cls.assert_not_called()

    @mock.patch("airflow.providers.slack.transfers.sql_to_slack.SlackHook")
    @mock.patch("airflow.providers.slack.transfers.sql_to_slack.BaseSqlToSlackOperator._get_query_results")
    def test_null_output_raise_error(self, mock_get_query_results, mock_slack_hook_cls):
        op_kwargs = {
            **self.default_op_kwargs,
            "slack_conn_id": "expected-test-slack-conn-id",
            "slack_filename": "test_filename.csv",
            "slack_channels": ["#random"],
            "slack_initial_comment": "test_comment",
            "slack_title": "test_title",
            "action_on_empty_df": "error",
        }
        op = SqlToSlackApiFileOperator(task_id="test_send_file", **op_kwargs)

        # Mock empty query results
        mock_df = mock.MagicMock()
        mock_df.configure_mock(**{"empty.return_value": True})
        mock_get_query_results.return_value = mock_df

        with pytest.raises(ValueError, match=r"output df must be non-empty\. Failing"):
            op.execute(mock.MagicMock())
        mock_slack_hook_cls.assert_not_called()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

from airflow.providers.common.compat.sdk import AirflowException, BaseHook, BaseOperator

if TYPE_CHECKING:
    import pandas as pd
    from slack_sdk.http_retry import RetryHandler

    from airflow.providers.common.sql.hooks.sql import DbApiHook


class BaseSqlToSlackOperator(BaseOperator):
    """
    Operator implements base sql methods for SQL to Slack Transfer operators.

    :param sql: The SQL query to be executed
    :param sql_conn_id: reference to a specific DB-API Connection.
    :param sql_hook_params: Extra config params to be passed to the underlying hook.
        Should match the desired hook constructor params.
    :param parameters: The parameters to pass to the SQL query.
    :param slack_proxy: Proxy to make the Slack Incoming Webhook / API calls. Optional
    :param slack_timeout: The maximum number of seconds the client will wait to connect
        and receive a response from Slack. Optional
    :param slack_retry_handlers: List of handlers to customize retry logic. Optional
    """

    def __init__(
        self,
        *,
        sql: str,
        sql_conn_id: str,
        sql_hook_params: dict | None = None,
        parameters: list | tuple | Mapping[str, Any] | None = None,
        slack_proxy: str | None = None,
        slack_timeout: int | None = None,
        slack_retry_handlers: list[RetryHandler] | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.sql_conn_id = sql_conn_id
        self.sql_hook_params = sql_hook_params
        self.sql = sql
        self.parameters = parameters
        self.slack_proxy = slack_proxy
        self.slack_timeout = slack_timeout
        self.slack_retry_handlers = slack_retry_handlers

    def _get_hook(self) -> DbApiHook:
        self.log.debug("Get connection for %s", self.sql_conn_id)
        conn = BaseHook.get_connection(self.sql_conn_id)
        hook = conn.get_hook(hook_params=self.sql_hook_params)
        if not callable(getattr(hook, "get_df", None)):
            raise AirflowException("This hook is not supported. The hook class must have get_df method.")
        return hook

    def _get_query_results(self) -> pd.DataFrame:
        sql_hook = self._get_hook()

        self.log.info("Running SQL query: %s", self.sql)
        df = sql_hook.get_df(self.sql, parameters=self.parameters)
        return df



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from unittest import mock

import pytest

from airflow.providers.slack.notifications.slack_webhook import (
    SlackWebhookNotifier,
    send_slack_webhook_notification,
)

DEFAULT_HOOKS_PARAMETERS = {"timeout": None, "proxy": None, "retry_handlers": None}


class TestSlackNotifier:
    def test_class_and_notifier_are_same(self):
        assert send_slack_webhook_notification is SlackWebhookNotifier

    @mock.patch("airflow.providers.slack.notifications.slack_webhook.SlackWebhookHook")
    @pytest.mark.parametrize(
        ("slack_op_kwargs", "hook_extra_kwargs"),
        [
            pytest.param({}, DEFAULT_HOOKS_PARAMETERS, id="default-hook-parameters"),
            pytest.param(
                {"timeout": 42, "proxy": "http://spam.egg", "retry_handlers": []},
                {"timeout": 42, "proxy": "http://spam.egg", "retry_handlers": []},
                id="with-extra-hook-parameters",
            ),
        ],
    )
    def test_slack_webhook_notifier(self, mock_slack_hook, slack_op_kwargs, hook_extra_kwargs):
        notifier = send_slack_webhook_notification(
            slack_webhook_conn_id="test_conn_id",
            text="foo-bar",
            blocks="spam-egg",
            attachments="baz-qux",
            unfurl_links=True,
            unfurl_media=False,
            **slack_op_kwargs,
        )
        notifier.notify({})
        mock_slack_hook.return_value.send.assert_called_once_with(
            text="foo-bar",
            blocks="spam-egg",
            unfurl_links=True,
            unfurl_media=False,
            attachments="baz-qux",
        )
        mock_slack_hook.assert_called_once_with(slack_webhook_conn_id="test_conn_id", **hook_extra_kwargs)

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.slack.notifications.slack_webhook.SlackWebhookHook")
    async def test_async_slack_webhook_notifier(self, mock_slack_hook):
        mock_hook = mock_slack_hook.return_value
        mock_hook.async_send = mock.AsyncMock()

        notifier = send_slack_webhook_notification(
            slack_webhook_conn_id="test_conn_id",
            text="foo-bar",
            blocks="spam-egg",
            attachments="baz-qux",
            unfurl_links=True,
            unfurl_media=False,
        )

        await notifier.async_notify({})

        mock_hook.async_send.assert_called_once_with(
            text="foo-bar",
            blocks="spam-egg",
            unfurl_links=True,
            unfurl_media=False,
            attachments="baz-qux",
        )

    @mock.patch("airflow.providers.slack.notifications.slack_webhook.SlackWebhookHook")
    def test_slack_webhook_templated(self, mock_slack_hook, create_dag_without_db):
        notifier = send_slack_webhook_notification(
            text="Who am I? {{ username }}",
            blocks=[{"type": "header", "text": {"type": "plain_text", "text": "{{ dag.dag_id }}"}}],
            attachments=[{"image_url": "{{ dag.dag_id }}.png"}],
        )
        notifier(
            {
                "dag": create_dag_without_db("test_send_slack_webhook_notification_templated"),
                "username": "not-a-root",
            }
        )
        mock_slack_hook.return_value.send.assert_called_once_with(
            text="Who am I? not-a-root",
            blocks=[
                {
                    "type": "header",
                    "text": {"type": "plain_text", "text": "test_send_slack_webhook_notification_templated"},
                }
            ],
            attachments=[{"image_url": "test_send_slack_webhook_notification_templated.png"}],
            unfurl_links=None,
            unfurl_media=None,
        )

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.slack.notifications.slack_webhook.SlackWebhookHook")
    async def test_async_slack_webhook_templated(self, mock_slack_hook, create_dag_without_db):
        """Test async notification with template rendering."""
        mock_hook = mock_slack_hook.return_value
        mock_hook.async_send = mock.AsyncMock()

        notifier = send_slack_webhook_notification(
            text="Who am I? {{ username }}",
            blocks=[{"type": "header", "text": {"type": "plain_text", "text": "{{ dag.dag_id }}"}}],
            attachments=[{"image_url": "{{ dag.dag_id }}.png"}],
        )

        # Call notifier first to handle template rendering
        notifier(
            {
                "dag": create_dag_without_db("test_async_send_slack_webhook_notification_templated"),
                "username": "not-a-root",
            }
        )

        # Then call async_notify with rendered templates
        await notifier.async_notify(
            {
                "dag": create_dag_without_db("test_async_send_slack_webhook_notification_templated"),
                "username": "not-a-root",
            }
        )

        mock_hook.async_send.assert_called_once_with(
            text="Who am I? not-a-root",
            blocks=[
                {
                    "type": "header",
                    "text": {
                        "type": "plain_text",
                        "text": "test_async_send_slack_webhook_notification_templated",
                    },
                }
            ],
            attachments=[{"image_url": "test_async_send_slack_webhook_notification_templated.png"}],
            unfurl_links=None,
            unfurl_media=None,
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pandas as pd
import pytest

from airflow.models import Connection
from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.slack.transfers.base_sql_to_slack import BaseSqlToSlackOperator


class TestBaseSqlToSlackOperator:
    def setup_method(self):
        self.default_op_kwargs = {
            "sql": "SELECT 1",
            "sql_conn_id": "test-sql-conn-id",
            "sql_hook_params": None,
            "parameters": None,
        }

    def test_execute_not_implemented(self):
        """Test that no base implementation for ``BaseSqlToSlackOperator.execute()``."""
        op = BaseSqlToSlackOperator(task_id="test_base_not_implements", **self.default_op_kwargs)
        with pytest.raises(NotImplementedError):
            op.execute(mock.MagicMock())

    @mock.patch("airflow.providers.common.sql.operators.sql.BaseHook.get_connection")
    @mock.patch("airflow.models.connection.Connection.get_hook")
    @pytest.mark.parametrize("conn_type", ["postgres", "snowflake"])
    @pytest.mark.parametrize("sql_hook_params", [None, {"foo": "bar"}])
    def test_get_hook(self, mock_get_hook, mock_get_conn, conn_type, sql_hook_params):
        class SomeDummyHook:
            """Hook which implements ``get_df`` method"""

            def get_df(self):
                pass

        expected_hook = SomeDummyHook()
        mock_get_conn.return_value = Connection(conn_id=f"test_connection_{conn_type}", conn_type=conn_type)
        mock_get_hook.return_value = expected_hook
        op_kwargs = {
            **self.default_op_kwargs,
            "sql_hook_params": sql_hook_params,
        }
        op = BaseSqlToSlackOperator(task_id="test_get_hook", **op_kwargs)
        hook = op._get_hook()
        mock_get_hook.assert_called_once_with(hook_params=sql_hook_params)
        assert hook == expected_hook

    @mock.patch("airflow.providers.common.sql.operators.sql.BaseHook.get_connection")
    @mock.patch("airflow.models.connection.Connection.get_hook")
    def test_get_not_supported_hook(self, mock_get_hook, mock_get_conn):
        class SomeDummyHook:
            """Hook which not implemented ``get_df`` method"""

        mock_get_conn.return_value = Connection(conn_id="test_connection", conn_type="test_connection")
        mock_get_hook.return_value = SomeDummyHook()
        op = BaseSqlToSlackOperator(task_id="test_get_not_supported_hook", **self.default_op_kwargs)
        error_message = r"This hook is not supported. The hook class must have get_df method\."
        with pytest.raises(AirflowException, match=error_message):
            op._get_hook()

    @mock.patch("airflow.providers.slack.transfers.sql_to_slack.BaseSqlToSlackOperator._get_hook")
    @pytest.mark.parametrize("sql", ["SELECT 42", "SELECT 1 FROM DUMMY WHERE col = ?"])
    @pytest.mark.parametrize("parameters", [None, {"col": "spam-egg"}])
    def test_get_query_results(self, mock_op_get_hook, sql, parameters):
        test_df = pd.DataFrame({"a": "1", "b": "2"}, index=[0, 1])
        mock_get_df = mock.MagicMock(return_value=test_df)
        mock_hook = mock.MagicMock()
        mock_hook.get_df = mock_get_df
        mock_op_get_hook.return_value = mock_hook
        op_kwargs = {
            **self.default_op_kwargs,
            "sql": sql,
            "parameters": parameters,
        }
        op = BaseSqlToSlackOperator(task_id="test_get_query_results", **op_kwargs)
        df = op._get_query_results()
        mock_get_df.assert_called_once_with(sql, parameters=parameters)
        assert df is test_df



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.providers.slack.operators.slack_webhook import SlackWebhookOperator

DEFAULT_HOOKS_PARAMETERS = {"timeout": None, "proxy": None, "retry_handlers": None}


class TestSlackWebhookOperator:
    def setup_method(self):
        self.default_op_kwargs = {
            "slack_webhook_conn_id": "test_conn_id",
            "channel": None,
            "username": None,
            "icon_emoji": None,
            "icon_url": None,
        }

    @mock.patch("airflow.providers.slack.operators.slack_webhook.SlackWebhookHook")
    @pytest.mark.parametrize(
        ("slack_op_kwargs", "hook_extra_kwargs"),
        [
            pytest.param({}, DEFAULT_HOOKS_PARAMETERS, id="default-hook-parameters"),
            pytest.param(
                {"timeout": 42, "proxy": "http://spam.egg", "retry_handlers": []},
                {"timeout": 42, "proxy": "http://spam.egg", "retry_handlers": []},
                id="with-extra-hook-parameters",
            ),
        ],
    )
    def test_hook(self, mock_slackwebhook_cls, slack_op_kwargs, hook_extra_kwargs):
        """Test get cached ``SlackWebhookHook`` hook."""
        op = SlackWebhookOperator(
            task_id="test_hook", slack_webhook_conn_id="test_conn_id", **slack_op_kwargs
        )
        hook = op.hook
        assert hook is op.hook, "Expected cached hook"
        mock_slackwebhook_cls.assert_called_once_with(
            slack_webhook_conn_id="test_conn_id", **hook_extra_kwargs
        )

    def test_assert_templated_fields(self):
        """Test expected templated fields."""
        operator = SlackWebhookOperator(task_id="test_assert_templated_fields", **self.default_op_kwargs)
        template_fields = (
            "message",
            "attachments",
            "blocks",
            "channel",
            "username",
            "proxy",
        )
        assert operator.template_fields == template_fields

    @pytest.mark.parametrize(
        ("message", "blocks", "attachments"),
        [
            ("Test Text", ["Dummy Block"], ["Test Attachments"]),
            ("Test Text", ["Dummy Block"], None),
            ("Test Text", None, None),
            (None, ["Dummy Block"], None),
            (None, ["Dummy Block"], ["Test Attachments"]),
            (None, None, ["Test Attachments"]),
        ],
    )
    @pytest.mark.parametrize(
        ("channel", "username", "icon_emoji", "icon_url"),
        [
            (None, None, None, None),
            ("legacy-channel", "legacy-username", "legacy-icon_emoji", "legacy-icon-url"),
        ],
        ids=["webhook-attrs", "legacy-webhook-attrs"],
    )
    @mock.patch("airflow.providers.slack.operators.slack_webhook.SlackWebhookHook")
    def test_execute_operator(
        self, mock_slackwebhook_cls, message, blocks, attachments, channel, username, icon_emoji, icon_url
    ):
        mock_slackwebhook = mock_slackwebhook_cls.return_value
        mock_slackwebhook_send = mock_slackwebhook.send
        op = SlackWebhookOperator(
            task_id="test_execute",
            slack_webhook_conn_id="test_conn_id",
            message=message,
            blocks=blocks,
            attachments=attachments,
            channel=channel,
            username=username,
            icon_emoji=icon_emoji,
            icon_url=icon_url,
        )
        op.execute(mock.MagicMock())
        mock_slackwebhook_send.assert_called_once_with(
            text=message,
            blocks=blocks,
            attachments=attachments,
            channel=channel,
            username=username,
            icon_emoji=icon_emoji,
            icon_url=icon_url,
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pandas as pd
import pytest

from airflow.models import Connection
from airflow.providers.slack.transfers.sql_to_slack_webhook import SqlToSlackWebhookOperator

from tests_common.test_utils.compat import timezone

TEST_DAG_ID = "sql_to_slack_unit_test"
TEST_TASK_ID = "sql_to_slack_unit_test_task"
DEFAULT_DATE = timezone.datetime(2017, 1, 1)


@pytest.fixture
def mocked_hook():
    with mock.patch("airflow.providers.slack.transfers.sql_to_slack_webhook.SlackWebhookHook") as m:
        yield m


class TestSqlToSlackWebhookOperator:
    @pytest.fixture(autouse=True)
    def setup_connections(self, create_connection_without_db):
        create_connection_without_db(
            Connection(
                conn_id="slack_connection",
                conn_type="slackwebhook",
                password="xoxb-1234567890123-09876543210987-AbCdEfGhIjKlMnOpQrStUvWx",
            )
        )

    def setup_method(self):
        self.default_hook_parameters = {"timeout": None, "proxy": None, "retry_handlers": None}

    @staticmethod
    def _construct_operator(**kwargs):
        operator = SqlToSlackWebhookOperator(task_id=TEST_TASK_ID, **kwargs)
        return operator

    @pytest.mark.parametrize(
        ("slack_op_kwargs", "hook_extra_kwargs"),
        [
            pytest.param(
                {}, {"timeout": None, "proxy": None, "retry_handlers": None}, id="default-hook-parameters"
            ),
            pytest.param(
                {"slack_timeout": 42, "slack_proxy": "http://spam.egg", "slack_retry_handlers": []},
                {"timeout": 42, "proxy": "http://spam.egg", "retry_handlers": []},
                id="with-extra-hook-parameters",
            ),
        ],
    )
    def test_rendering_and_message_execution(self, slack_op_kwargs, hook_extra_kwargs, mocked_hook):
        mock_dbapi_hook = mock.Mock()

        test_df = pd.DataFrame({"a": "1", "b": "2"}, index=[0, 1])
        get_df_mock = mock_dbapi_hook.return_value.get_df
        get_df_mock.return_value = test_df

        operator_args = {
            "sql_conn_id": "snowflake_connection",
            "slack_webhook_conn_id": "slack_connection",
            "slack_message": "message: {{ ds }}, {{ results_df }}",
            "slack_channel": "#test",
            "sql": "sql {{ ds }}",
            **slack_op_kwargs,
        }
        sql_to_slack_operator = self._construct_operator(**operator_args)

        slack_webhook_hook = mocked_hook.return_value
        sql_to_slack_operator._get_hook = mock_dbapi_hook
        sql_to_slack_operator.render_template_fields({"ds": "2017-01-01"})
        sql_to_slack_operator.execute({"ds": "2017-01-01"})

        # Test that the Slack hook is instantiated with the right parameters
        mocked_hook.assert_called_once_with(slack_webhook_conn_id="slack_connection", **hook_extra_kwargs)

        # Test that the `SlackWebhookHook.send` method gets run once
        slack_webhook_hook.send.assert_called_once_with(
            text=f"message: 2017-01-01, {test_df}",
            channel="#test",
        )

    def test_rendering_and_message_execution_with_slack_hook(self, mocked_hook):
        mock_dbapi_hook = mock.Mock()

        test_df = pd.DataFrame({"a": "1", "b": "2"}, index=[0, 1])
        get_df_mock = mock_dbapi_hook.return_value.get_df
        get_df_mock.return_value = test_df

        operator_args = {
            "sql_conn_id": "snowflake_connection",
            "slack_webhook_conn_id": "slack_connection",
            "slack_message": "message: {{ ds }}, {{ results_df }}",
            "slack_channel": "#test",
            "sql": "sql {{ ds }}",
        }
        sql_to_slack_operator = self._construct_operator(**operator_args)

        slack_webhook_hook = mocked_hook.return_value
        sql_to_slack_operator._get_hook = mock_dbapi_hook

        sql_to_slack_operator.render_template_fields({"ds": "2017-01-01"})
        sql_to_slack_operator.execute({"ds": "2017-01-01"})

        # Test that the Slack hook is instantiated with the right parameters
        mocked_hook.assert_called_once_with(
            slack_webhook_conn_id="slack_connection", **self.default_hook_parameters
        )

        # Test that the `SlackWebhookHook.send` method gets run once
        slack_webhook_hook.send.assert_called_once_with(
            text=f"message: 2017-01-01, {test_df}",
            channel="#test",
        )

    @pytest.mark.parametrize(
        ("slack_webhook_conn_id", "warning_expected", "expected_conn_id"),
        [
            pytest.param("foo", False, "foo", id="slack-webhook-conn-id"),
            pytest.param("spam", True, "spam", id="mixin-conn-ids"),
        ],
    )
    def test_resolve_conn_ids(self, slack_webhook_conn_id, warning_expected, expected_conn_id):
        operator_args = {
            "sql_conn_id": "snowflake_connection",
            "slack_message": "message: {{ ds }}, {{ xxxx }}",
            "sql": "sql {{ ds }}",
        }
        if slack_webhook_conn_id:
            operator_args["slack_webhook_conn_id"] = slack_webhook_conn_id

        op = self._construct_operator(**operator_args)

        assert op.slack_webhook_conn_id == expected_conn_id

    def test_non_existing_slack_webhook_conn_id(self):
        operator_args = {
            "sql_conn_id": "snowflake_connection",
            "slack_message": "message: {{ ds }}, {{ xxxx }}",
            "sql": "sql {{ ds }}",
        }
        with pytest.raises(ValueError, match="Got an empty `slack_webhook_conn_id` value"):
            self._construct_operator(**operator_args)

    def test_rendering_custom_df_name_message_execution(self, mocked_hook):
        mock_dbapi_hook = mock.Mock()

        test_df = pd.DataFrame({"a": "1", "b": "2"}, index=[0, 1])
        get_df_mock = mock_dbapi_hook.return_value.get_df
        get_df_mock.return_value = test_df

        operator_args = {
            "sql_conn_id": "snowflake_connection",
            "slack_webhook_conn_id": "slack_connection",
            "slack_message": "message: {{ ds }}, {{ testing }}",
            "slack_channel": "#test",
            "sql": "sql {{ ds }}",
            "results_df_name": "testing",
        }
        sql_to_slack_operator = self._construct_operator(**operator_args)

        slack_webhook_hook = mocked_hook.return_value
        sql_to_slack_operator._get_hook = mock_dbapi_hook

        sql_to_slack_operator.render_template_fields({"ds": "2017-01-01"})
        sql_to_slack_operator.execute({"ds": "2017-01-01"})

        # Test that the Slack hook is instantiated with the right parameters
        mocked_hook.assert_called_once_with(
            slack_webhook_conn_id="slack_connection", **self.default_hook_parameters
        )

        # Test that the `SlackWebhookHook.send` method gets run once
        slack_webhook_hook.send.assert_called_once_with(
            text=f"message: 2017-01-01, {test_df}",
            channel="#test",
        )

    def test_hook_params_building(self, mocked_get_connection):
        mocked_get_connection.return_value = Connection(conn_id="snowflake_connection", conn_type="snowflake")
        hook_params = {
            "schema": "test_schema",
            "role": "test_role",
            "database": "test_database",
            "warehouse": "test_warehouse",
        }
        operator_args = {
            "sql_conn_id": "dummy_connection",
            "sql": "sql {{ ds }}",
            "results_df_name": "xxxx",
            "sql_hook_params": hook_params,
            "slack_webhook_conn_id": "slack_connection",
            "parameters": ["1", "2", "3"],
            "slack_message": "message: {{ ds }}, {{ xxxx }}",
        }
        sql_to_slack_operator = SqlToSlackWebhookOperator(task_id=TEST_TASK_ID, **operator_args)

        assert sql_to_slack_operator.sql_hook_params == hook_params

    def test_hook_params(self, mocked_get_connection):
        mocked_get_connection.return_value = Connection(conn_id="postgres_test", conn_type="postgres")
        op = SqlToSlackWebhookOperator(
            task_id="sql_hook_params",
            sql_conn_id="postgres_test",
            slack_webhook_conn_id="slack_connection",
            sql="SELECT 1",
            slack_message="message: {{ ds }}, {{ xxxx }}",
            sql_hook_params={
                "log_sql": False,
            },
        )
        hook = op._get_hook()
        assert hook.log_sql == op.sql_hook_params["log_sql"]

    def test_hook_params_snowflake(self, mocked_get_connection):
        mocked_get_connection.return_value = Connection(conn_id="snowflake_default", conn_type="snowflake")
        op = SqlToSlackWebhookOperator(
            task_id="snowflake_hook_params",
            sql_conn_id="snowflake_default",
            slack_webhook_conn_id="slack_default",
            results_df_name="xxxx",
            sql="SELECT 1",
            slack_message="message: {{ ds }}, {{ xxxx }}",
            sql_hook_params={
                "warehouse": "warehouse",
                "database": "database",
                "role": "role",
                "schema": "schema",
            },
        )
        hook = op._get_hook()

        assert hook.warehouse == "warehouse"
        assert hook.database == "database"
        assert hook.role == "role"
        assert hook.schema == "schema"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from unittest import mock

import pytest

from airflow.providers.slack.notifications.slack import SlackNotifier, send_slack_notification

DEFAULT_HOOKS_PARAMETERS = {"base_url": None, "timeout": None, "proxy": None, "retry_handlers": None}


class TestSlackNotifier:
    @mock.patch("airflow.providers.slack.notifications.slack.SlackHook")
    @pytest.mark.parametrize(
        ("extra_kwargs", "hook_extra_kwargs"),
        [
            pytest.param({}, DEFAULT_HOOKS_PARAMETERS, id="default-hook-parameters"),
            pytest.param(
                {
                    "base_url": "https://foo.bar",
                    "timeout": 42,
                    "proxy": "http://spam.egg",
                    "retry_handlers": [],
                },
                {
                    "base_url": "https://foo.bar",
                    "timeout": 42,
                    "proxy": "http://spam.egg",
                    "retry_handlers": [],
                },
                id="with-extra-hook-parameters",
            ),
        ],
    )
    def test_slack_notifier(self, mock_slack_hook, create_dag_without_db, extra_kwargs, hook_extra_kwargs):
        notifier = send_slack_notification(slack_conn_id="test_conn_id", text="test", **extra_kwargs)
        notifier({"dag": create_dag_without_db("test_slack_notifier")})
        mock_slack_hook.return_value.call.assert_called_once_with(
            "chat.postMessage",
            json={
                "channel": "#general",
                "username": "Airflow",
                "text": "test",
                "icon_url": "https://raw.githubusercontent.com/apache/airflow/main/airflow-core"
                "/src/airflow/ui/public/pin_100.png",
                "attachments": "[]",
                "blocks": "[]",
                "unfurl_links": True,
                "unfurl_media": True,
            },
        )
        mock_slack_hook.assert_called_once_with(slack_conn_id="test_conn_id", **hook_extra_kwargs)

    @mock.patch("airflow.providers.slack.notifications.slack.SlackHook")
    def test_slack_notifier_with_notifier_class(self, mock_slack_hook, create_dag_without_db):
        notifier = SlackNotifier(text="test")
        notifier({"dag": create_dag_without_db("test_slack_notifier")})
        mock_slack_hook.return_value.call.assert_called_once_with(
            "chat.postMessage",
            json={
                "channel": "#general",
                "username": "Airflow",
                "text": "test",
                "icon_url": "https://raw.githubusercontent.com/apache/airflow/main/airflow-core"
                "/src/airflow/ui/public/pin_100.png",
                "attachments": "[]",
                "blocks": "[]",
                "unfurl_links": True,
                "unfurl_media": True,
            },
        )

    @mock.patch("airflow.providers.slack.notifications.slack.SlackHook")
    def test_slack_notifier_templated(self, mock_slack_hook, create_dag_without_db):
        notifier = send_slack_notification(
            text="test {{ username }}",
            channel="#test-{{dag.dag_id}}",
            attachments=[{"image_url": "{{ dag.dag_id }}.png"}],
        )
        context = {"dag": create_dag_without_db("test_slack_notifier")}
        notifier(context)
        mock_slack_hook.return_value.call.assert_called_once_with(
            "chat.postMessage",
            json={
                "channel": "#test-test_slack_notifier",
                "username": "Airflow",
                "text": "test Airflow",
                "icon_url": "https://raw.githubusercontent.com/apache/airflow/main/airflow-core"
                "/src/airflow/ui/public/pin_100.png",
                "attachments": '[{"image_url": "test_slack_notifier.png"}]',
                "blocks": "[]",
                "unfurl_links": True,
                "unfurl_media": True,
            },
        )

    @mock.patch("airflow.providers.slack.notifications.slack.SlackHook")
    def test_slack_notifier_unfurl_options(self, mock_slack_hook, create_dag_without_db):
        notifier = send_slack_notification(
            text="test",
            unfurl_links=False,
            unfurl_media=False,
        )
        notifier({"dag": create_dag_without_db("test_slack_notifier")})
        mock_slack_hook.return_value.call.assert_called_once_with(
            "chat.postMessage",
            json={
                "channel": "#general",
                "username": "Airflow",
                "text": "test",
                "icon_url": "https://raw.githubusercontent.com/apache/airflow/main/airflow-core"
                "/src/airflow/ui/public/pin_100.png",
                "attachments": "[]",
                "blocks": "[]",
                "unfurl_links": False,
                "unfurl_media": False,
            },
        )

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.slack.notifications.slack.SlackHook")
    async def test_async_slack_notifier(self, mock_slack_hook):
        mock_slack_hook.return_value.async_call = mock.AsyncMock()

        notifier = send_slack_notification(
            text="test",
            unfurl_links=False,
            unfurl_media=False,
        )

        await notifier.async_notify({})

        mock_slack_hook.return_value.async_call.assert_called_once_with(
            "chat.postMessage",
            json={
                "channel": "#general",
                "username": "Airflow",
                "text": "test",
                "icon_url": "https://raw.githubusercontent.com/apache/airflow/main/airflow-core"
                "/src/airflow/ui/public/pin_100.png",
                "attachments": "[]",
                "blocks": "[]",
                "unfurl_links": False,
                "unfurl_media": False,
            },
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from typing import Any
from unittest import mock

import pytest
from slack_sdk.errors import SlackApiError
from slack_sdk.http_retry.builtin_handlers import ConnectionErrorRetryHandler, RateLimitErrorRetryHandler
from slack_sdk.web.slack_response import SlackResponse

from airflow.models.connection import Connection
from airflow.providers.common.compat.sdk import AirflowException, AirflowNotFoundException
from airflow.providers.slack.hooks.slack import SlackHook

MOCK_SLACK_API_TOKEN = "xoxb-1234567890123-09876543210987-AbCdEfGhIjKlMnOpQrStUvWx"
SLACK_API_DEFAULT_CONN_ID = SlackHook.default_conn_name
SLACK_CONN_TYPE = "slack"
TEST_CONN_ERROR_RETRY_HANDLER = ConnectionErrorRetryHandler(max_retry_count=42)
TEST_RATE_LIMIT_RETRY_HANDLER = RateLimitErrorRetryHandler()

CONN_TYPE = "slack-incoming-webhook"
VALID_CONN_IDS = [
    SlackHook.default_conn_name,
    "conn_full_url_connection",
    "conn_host_with_schema",
    "conn_parts",
]


@pytest.fixture(scope="module", autouse=True)
def slack_api_connections():
    """Create tests connections."""
    connections = [
        Connection(
            conn_id=SLACK_API_DEFAULT_CONN_ID,
            conn_type=CONN_TYPE,
            password=MOCK_SLACK_API_TOKEN,
        ),
        Connection(
            conn_id="compat_http_type",
            conn_type="http",
            password=MOCK_SLACK_API_TOKEN,
        ),
        Connection(
            conn_id="empty_slack_connection",
            conn_type=CONN_TYPE,
        ),
    ]

    with pytest.MonkeyPatch.context() as mp:
        for conn in connections:
            mp.setenv(f"AIRFLOW_CONN_{conn.conn_id.upper()}", conn.get_uri())
        yield


@pytest.fixture
def mocked_client():
    with mock.patch.object(SlackHook, "client") as m:
        yield m


class TestSlackHook:
    @staticmethod
    def fake_slack_response(*, data: dict | bytes, status_code: int = 200, **kwargs):
        """Helper for generate fake slack response."""
        # Mock other mandatory ``SlackResponse`` arguments
        for mandatory_param in ("client", "http_verb", "api_url", "req_args", "headers"):
            if mandatory_param not in kwargs:
                kwargs[mandatory_param] = mock.MagicMock(name=f"fake-{mandatory_param}")

        return SlackResponse(status_code=status_code, data=data, **kwargs)

    @staticmethod
    def make_429():
        resp = mock.MagicMock()
        resp.status_code = 429
        resp.headers = {"Retry-After": "1"}
        return SlackApiError("ratelimited", response=resp)

    @pytest.mark.parametrize(
        "conn_id",
        [
            SLACK_API_DEFAULT_CONN_ID,
            "compat_http_type",  # In case if users use Slack Webhook Connection ID from UI for Slack API
        ],
    )
    def test_get_token_from_connection(self, conn_id):
        """Test retrieve token from Slack API Connection ID."""
        hook = SlackHook(slack_conn_id=SLACK_API_DEFAULT_CONN_ID)
        conn = hook.get_connection(hook.slack_conn_id)
        assert hook._get_conn_params(conn)["token"] == MOCK_SLACK_API_TOKEN

    def test_resolve_token(self):
        """Test that we only use token from Slack API Connection ID."""
        with pytest.warns(UserWarning, match="Provide `token` as part of .* parameters is disallowed"):
            hook = SlackHook(slack_conn_id=SLACK_API_DEFAULT_CONN_ID, token="foo-bar")
        conn = hook.get_connection(hook.slack_conn_id)
        assert "token" not in hook.extra_client_args
        assert hook._get_conn_params(conn)["token"] == MOCK_SLACK_API_TOKEN

    def test_empty_password(self):
        """Test password field defined in the connection."""
        hook = SlackHook(slack_conn_id="empty_slack_connection")
        conn = hook.get_connection(hook.slack_conn_id)
        error_message = r"Connection ID '.*' does not contain password \(Slack API Token\)\."
        with pytest.raises(AirflowNotFoundException, match=error_message):
            hook._get_conn_params(conn)

    @pytest.mark.parametrize(
        ("hook_config", "conn_extra", "expected"),
        [
            (  # Test Case: hook config
                {
                    "timeout": 42,
                    "base_url": "http://hook-base-url:1234",
                    "proxy": "https://hook-proxy:1234",
                    "retry_handlers": [TEST_CONN_ERROR_RETRY_HANDLER, TEST_RATE_LIMIT_RETRY_HANDLER],
                },
                {},
                {
                    "timeout": 42,
                    "base_url": "http://hook-base-url:1234",
                    "proxy": "https://hook-proxy:1234",
                    "retry_handlers": [TEST_CONN_ERROR_RETRY_HANDLER, TEST_RATE_LIMIT_RETRY_HANDLER],
                },
            ),
            (  # Test Case: connection config
                {},
                {
                    "timeout": "9000",
                    "base_url": "http://conn-base-url:4321",
                    "proxy": "https://conn-proxy:4321",
                },
                {
                    "timeout": 9000,
                    "base_url": "http://conn-base-url:4321",
                    "proxy": "https://conn-proxy:4321",
                },
            ),
            (  # Test Case: Connection from the UI
                {},
                {
                    "extra__slack__timeout": 9000,
                    "extra__slack__base_url": "http://conn-base-url:4321",
                    "extra__slack__proxy": "https://conn-proxy:4321",
                },
                {
                    "timeout": 9000,
                    "base_url": "http://conn-base-url:4321",
                    "proxy": "https://conn-proxy:4321",
                },
            ),
            (  # Test Case: Merge configs - hook args overwrite conn config
                {
                    "timeout": 1,
                    "proxy": "https://hook-proxy:777",
                    "retry_handlers": [TEST_RATE_LIMIT_RETRY_HANDLER],
                },
                {
                    "timeout": 9000,
                    "proxy": "https://conn-proxy:4321",
                },
                {
                    "timeout": 1,
                    "proxy": "https://hook-proxy:777",
                    "retry_handlers": [TEST_RATE_LIMIT_RETRY_HANDLER],
                },
            ),
            (  # Test Case: Merge configs - resolve config
                {
                    "timeout": 1,
                },
                {
                    "timeout": 9000,
                    "proxy": "https://conn-proxy:4334",
                },
                {
                    "timeout": 1,
                    "proxy": "https://conn-proxy:4334",
                },
            ),
            (  # Test Case: empty configs
                {},
                {},
                {},
            ),
            (  # Test Case: extra_client_args
                {"foo": "bar"},
                {},
                {"foo": "bar"},
            ),
            (  # Test Case: ignored not expected connection extra
                {},
                {"spam": "egg"},
                {},
            ),
        ],
    )
    @mock.patch("airflow.providers.slack.hooks.slack.WebClient")
    def test_client_configuration(
        self, mock_webclient_cls, hook_config, conn_extra, expected: dict[str, Any]
    ):
        """Test read/parse/merge WebClient config from connection and hook arguments."""
        expected["token"] = MOCK_SLACK_API_TOKEN
        test_conn = Connection(
            conn_id="test-slack-conn",
            conn_type=CONN_TYPE,
            password=MOCK_SLACK_API_TOKEN,
            extra=conn_extra,
        )
        test_conn_env = f"AIRFLOW_CONN_{test_conn.conn_id.upper()}"
        mock_webclient = mock_webclient_cls.return_value

        with mock.patch.dict("os.environ", values={test_conn_env: test_conn.get_uri()}):
            hook = SlackHook(slack_conn_id=test_conn.conn_id, **hook_config)
            conn = hook.get_connection(hook.slack_conn_id)
            expected["logger"] = hook.log
            conn_params = hook._get_conn_params(conn)
            assert conn_params == expected

            client = hook.client
            assert client == mock_webclient
            assert hook.get_conn() == mock_webclient
            assert hook.get_conn() is client  # cached
            mock_webclient_cls.assert_called_once_with(**expected)

    def test_call_with_failure(self, mocked_client):
        mocked_client.api_call.side_effect = SlackApiError(message="foo", response="bar")

        slack_hook = SlackHook(slack_conn_id=SLACK_API_DEFAULT_CONN_ID)
        test_method = "test_method"
        test_api_params = {"key1": "value1", "key2": "value2"}

        with pytest.raises(SlackApiError):
            slack_hook.call(test_method, data=test_api_params)

    def test_api_call(self, mocked_client):
        mocked_client.api_call.return_value = {"ok": True}

        slack_hook = SlackHook(slack_conn_id=SLACK_API_DEFAULT_CONN_ID)
        test_api_json = {"channel": "test_channel"}

        slack_hook.call("chat.postMessage", json=test_api_json)
        mocked_client.api_call.assert_called_with("chat.postMessage", json=test_api_json)

    @pytest.mark.parametrize(
        "response_data",
        [
            {
                "ok": True,
                "url": "https://subarachnoid.slack.com/",
                "team": "Subarachnoid Workspace",
                "user": "grace",
                "team_id": "T12345678",
                "user_id": "W12345678",
            },
            {
                "ok": True,
                "url": "https://subarachnoid.slack.com/",
                "team": "Subarachnoid Workspace",
                "user": "bot",
                "team_id": "T0G9PQBBK",
                "user_id": "W23456789",
                "bot_id": "BZYBOTHED",
            },
            b"some-binary-data",
        ],
    )
    def test_hook_connection_success(self, mocked_client, response_data):
        """Test SlackHook success connection."""
        mocked_client.api_call.return_value = self.fake_slack_response(status_code=200, data=response_data)
        hook = SlackHook(slack_conn_id=SLACK_API_DEFAULT_CONN_ID)
        conn_test = hook.test_connection()
        mocked_client.api_call.assert_called_once_with("auth.test")
        assert conn_test[0]

    @pytest.mark.parametrize(
        "response_data",
        [
            {"ok": False, "error": "invalid_auth"},
            {"ok": False, "error": "not_authed"},
            b"some-binary-data",
        ],
    )
    def test_hook_connection_failed(self, mocked_client, response_data):
        """Test SlackHook failure connection."""
        mocked_client.api_call.return_value = self.fake_slack_response(status_code=401, data=response_data)
        hook = SlackHook(slack_conn_id=SLACK_API_DEFAULT_CONN_ID)
        conn_test = hook.test_connection()
        mocked_client.api_call.assert_called_once_with("auth.test")
        assert not conn_test[0]

    @pytest.mark.parametrize(
        "uri",
        [
            pytest.param(
                "a://:abc@?extra__slack__timeout=123"
                "&extra__slack__base_url=base_url"
                "&extra__slack__proxy=proxy",
                id="prefix",
            ),
            pytest.param("a://:abc@?timeout=123&base_url=base_url&proxy=proxy", id="no-prefix"),
        ],
    )
    def test_backcompat_prefix_works(self, uri, monkeypatch):
        monkeypatch.setenv("AIRFLOW_CONN_MY_CONN", uri)
        hook = SlackHook(slack_conn_id="my_conn")
        conn = hook.get_connection(hook.slack_conn_id)
        params = hook._get_conn_params(conn)
        assert params["token"] == "abc"
        assert params["timeout"] == 123
        assert params["base_url"] == "base_url"
        assert params["proxy"] == "proxy"

    def test_backcompat_prefix_both_causes_warning(self, monkeypatch):
        monkeypatch.setenv("AIRFLOW_CONN_MY_CONN", "a://:abc@?extra__slack__timeout=111&timeout=222")
        hook = SlackHook(slack_conn_id="my_conn")
        conn = hook.get_connection(hook.slack_conn_id)
        with pytest.warns(Warning, match="Using value for `timeout`"):
            params = hook._get_conn_params(conn)
        assert params["timeout"] == 222

    def test_empty_string_ignored_prefixed(self, monkeypatch):
        monkeypatch.setenv(
            "AIRFLOW_CONN_MY_CONN",
            json.dumps(
                {"password": "hi", "extra": {"extra__slack__base_url": "", "extra__slack__proxy": ""}}
            ),
        )
        hook = SlackHook(slack_conn_id="my_conn")
        conn = hook.get_connection(hook.slack_conn_id)
        params = hook._get_conn_params(conn)
        assert "proxy" not in params
        assert "base_url" not in params

    def test_empty_string_ignored_non_prefixed(self, monkeypatch):
        monkeypatch.setenv(
            "AIRFLOW_CONN_MY_CONN",
            json.dumps({"password": "hi", "extra": {"base_url": "", "proxy": ""}}),
        )
        hook = SlackHook(slack_conn_id="my_conn")
        conn = hook.get_connection(hook.slack_conn_id)
        params = hook._get_conn_params(conn)
        assert "proxy" not in params
        assert "base_url" not in params

    def test_default_conn_name(self):
        hook = SlackHook()
        assert hook.slack_conn_id == SlackHook.default_conn_name

    def test_get_channel_id(self, mocked_client):
        fake_responses = [
            self.fake_slack_response(
                data={
                    "channels": [
                        {"id": "C0000000000", "name": "development-first-pr-support"},
                        {"id": "C0000000001", "name": "development"},
                    ],
                    "response_metadata": {"next_cursor": "FAKE"},
                }
            ),
            self.fake_slack_response(data={"response_metadata": {"next_cursor": "FAKE"}}),
            self.fake_slack_response(data={"channels": [{"id": "C0000000002", "name": "random"}]}),
            # Below should not reach, because previous one doesn't contain ``next_cursor``
            self.fake_slack_response(data={"channels": [{"id": "C0000000003", "name": "troubleshooting"}]}),
        ]
        mocked_client.conversations_list.side_effect = fake_responses

        hook = SlackHook(slack_conn_id=SLACK_API_DEFAULT_CONN_ID)

        assert hook.get_channel_id("development") == "C0000000001"
        mocked_client.conversations_list.assert_called()

        mocked_client.conversations_list.reset_mock()
        mocked_client.conversations_list.side_effect = fake_responses
        assert hook.get_channel_id("development-first-pr-support") == "C0000000000"
        # It should use cached values, so there is no new calls expected here
        mocked_client.assert_not_called()

        # Test pagination
        mocked_client.conversations_list.side_effect = fake_responses
        assert hook.get_channel_id("random") == "C0000000002"

        # Test non-existed channels
        mocked_client.conversations_list.side_effect = fake_responses
        with pytest.raises(LookupError, match="Unable to find slack channel"):
            hook.get_channel_id("troubleshooting")

    def test_call_conversations_list_retries_then_succeeds(self, monkeypatch):
        ok_resp = self.fake_slack_response(data={"channels": []})
        monkeypatch.setattr(
            "airflow.providers.slack.hooks.slack.WebClient",
            lambda **_: mock.MagicMock(
                conversations_list=mock.Mock(side_effect=[self.make_429(), self.make_429(), ok_resp])
            ),
        )
        with mock.patch("time.sleep") as mocked_sleep:
            hook = SlackHook(slack_conn_id=SLACK_API_DEFAULT_CONN_ID)
            res = hook._call_conversations_list()
            assert res is ok_resp
            assert mocked_sleep.call_count == 2

    def test_call_conversations_list_exceeds_max(self, monkeypatch):
        monkeypatch.setattr(
            "airflow.providers.slack.hooks.slack.WebClient",
            lambda **_: mock.MagicMock(conversations_list=mock.Mock(side_effect=[self.make_429()] * 5)),
        )
        with pytest.raises(AirflowException, match="Max retries"):
            SlackHook(slack_conn_id=SLACK_API_DEFAULT_CONN_ID)._call_conversations_list()

    def test_send_file_v2(self, mocked_client):
        SlackHook(slack_conn_id=SLACK_API_DEFAULT_CONN_ID).send_file_v2(
            channel_id="C00000000", file_uploads={"file": "/foo/bar/file.txt", "filename": "foo.txt"}
        )
        mocked_client.files_upload_v2.assert_called_once_with(
            channel="C00000000",
            file_uploads=[{"file": "/foo/bar/file.txt", "filename": "foo.txt"}],
            initial_comment=None,
        )

    def test_send_file_v2_multiple_files(self, mocked_client):
        SlackHook(slack_conn_id=SLACK_API_DEFAULT_CONN_ID).send_file_v2(
            file_uploads=[
                {"file": "/foo/bar/file.txt"},
                {"content": "Some Text", "filename": "foo.txt"},
            ],
            initial_comment="Awesome File",
        )
        mocked_client.files_upload_v2.assert_called_once_with(
            channel=None,
            file_uploads=[
                {"file": "/foo/bar/file.txt", "filename": "Uploaded file"},
                {"content": "Some Text", "filename": "foo.txt"},
            ],
            initial_comment="Awesome File",
        )

    def test_send_file_v2_channel_name(self, mocked_client, caplog):
        with mock.patch.object(SlackHook, "get_channel_id", return_value="C00") as mocked_get_channel_id:
            warning_message = "consider replacing '#random' with the corresponding Channel ID 'C00'"
            with pytest.warns(UserWarning, match=warning_message):
                SlackHook(slack_conn_id=SLACK_API_DEFAULT_CONN_ID).send_file_v2(
                    channel_id="#random", file_uploads={"file": "/foo/bar/file.txt"}
                )
            mocked_get_channel_id.assert_called_once_with("random")
            mocked_client.files_upload_v2.assert_called_once_with(
                channel="C00",
                file_uploads=mock.ANY,
                initial_comment=mock.ANY,
            )

    @pytest.mark.parametrize("initial_comment", [None, "test comment"])
    @pytest.mark.parametrize("title", [None, "test title"])
    @pytest.mark.parametrize("filename", [None, "foo.bar"])
    @pytest.mark.parametrize("channel", [None, "#random"])
    @pytest.mark.parametrize("snippet_type", [None, "text"])
    def test_send_file_v1_to_v2_content(self, initial_comment, title, filename, channel, snippet_type):
        hook = SlackHook(slack_conn_id=SLACK_API_DEFAULT_CONN_ID)
        with mock.patch.object(SlackHook, "send_file_v2") as mocked_send_file_v2:
            hook.send_file_v1_to_v2(
                channels=channel,
                content='{"foo": "bar"}',
                filename=filename,
                initial_comment=initial_comment,
                title=title,
                snippet_type=snippet_type,
            )
            mocked_send_file_v2.assert_called_once_with(
                channel_id=channel,
                file_uploads={
                    "content": '{"foo": "bar"}',
                    "filename": filename,
                    "title": title,
                    "snippet_type": snippet_type,
                },
                initial_comment=initial_comment,
            )

    @pytest.mark.parametrize("initial_comment", [None, "test comment"])
    @pytest.mark.parametrize("title", [None, "test title"])
    @pytest.mark.parametrize("filename", [None, "foo.bar"])
    @pytest.mark.parametrize("channel", [None, "#random"])
    @pytest.mark.parametrize("snippet_type", [None, "text"])
    def test_send_file_v1_to_v2_file(self, initial_comment, title, filename, channel, snippet_type):
        hook = SlackHook(slack_conn_id=SLACK_API_DEFAULT_CONN_ID)
        with mock.patch.object(SlackHook, "send_file_v2") as mocked_send_file_v2:
            hook.send_file_v1_to_v2(
                channels=channel,
                file="/foo/bar/spam.egg",
                filename=filename,
                initial_comment=initial_comment,
                title=title,
                snippet_type=snippet_type,
            )
            mocked_send_file_v2.assert_called_once_with(
                channel_id=channel,
                file_uploads={
                    "file": "/foo/bar/spam.egg",
                    "filename": filename or "spam.egg",
                    "title": title,
                    "snippet_type": snippet_type,
                },
                initial_comment=initial_comment,
            )

    @pytest.mark.parametrize(
        ("file", "content"),
        [
            pytest.param(None, None, id="both-none"),
            pytest.param("", "", id="both-empty"),
            pytest.param("foo.bar", "test-content", id="both-specified"),
        ],
    )
    def test_send_file_v1_to_v2_wrong_parameters(self, file, content):
        hook = SlackHook(slack_conn_id=SLACK_API_DEFAULT_CONN_ID)
        error_message = "Either `file` or `content` must be provided, not both"
        with pytest.raises(ValueError, match=error_message):
            hook.send_file_v1_to_v2(file=file, content=content)

    @pytest.mark.parametrize(
        ("channels", "expected_calls"),
        [
            pytest.param("#foo, #bar", 2, id="comma-separated-string"),
            pytest.param(["#random", "#development", "#airflow-upgrades"], 3, id="list"),
        ],
    )
    def test_send_file_v1_to_v2_multiple_channels(self, channels, expected_calls):
        hook = SlackHook(slack_conn_id=SLACK_API_DEFAULT_CONN_ID)
        with mock.patch.object(SlackHook, "send_file_v2") as mocked_send_file_v2:
            hook.send_file_v1_to_v2(channels=channels, content="Fake")
            assert mocked_send_file_v2.call_count == expected_calls


class TestSlackHookAsync:
    @pytest.fixture
    def mock_get_conn(self):
        with mock.patch(
            "airflow.providers.slack.hooks.slack.get_async_connection", new_callable=mock.AsyncMock
        ) as m:
            m.return_value = Connection(
                conn_id=SLACK_API_DEFAULT_CONN_ID,
                conn_type=CONN_TYPE,
                password=MOCK_SLACK_API_TOKEN,
            )
            yield m

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.slack.hooks.slack.AsyncWebClient")
    async def test_get_async_client(self, mock_client, mock_get_conn):
        """Test get_async_client creates AsyncWebClient with correct params."""
        hook = SlackHook(slack_conn_id=SLACK_API_DEFAULT_CONN_ID)
        await hook.get_async_client()
        mock_get_conn.assert_called()
        mock_client.assert_called_once_with(token=MOCK_SLACK_API_TOKEN, logger=mock.ANY)

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.slack.hooks.slack.AsyncWebClient.api_call", new_callable=mock.AsyncMock)
    async def test_async_call(self, mock_api_call, mock_get_conn):
        """Test async_call is called correctly."""
        hook = SlackHook(slack_conn_id=SLACK_API_DEFAULT_CONN_ID)
        test_api_json = {"channel": "test_channel"}
        await hook.async_call("chat.postMessage", json=test_api_json)
        mock_get_conn.assert_called()
        mock_api_call.assert_called_with("chat.postMessage", json=test_api_json)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example DAG using SqlToSlackOperator.
"""

from __future__ import annotations

import os
from datetime import datetime

from airflow import models
from airflow.providers.slack.transfers.sql_to_slack_webhook import SqlToSlackWebhookOperator

SQL_TABLE = os.environ.get("SQL_TABLE", "test_table")
SQL_CONN_ID = "presto_default"
ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID")
DAG_ID = "example_sql_to_slack_webhook"

with models.DAG(
    dag_id=DAG_ID,
    schedule="@once",  # Override to match your needs
    start_date=datetime(2022, 1, 1),
    catchup=False,
    tags=["example"],
) as dag:
    # [START howto_operator_sql_to_slack_webhook]
    SqlToSlackWebhookOperator(
        task_id="presto_to_slack",
        sql_conn_id=SQL_CONN_ID,
        sql=f"SELECT col FROM {SQL_TABLE}",
        slack_channel="my_channel",
        slack_webhook_conn_id="slack_default",
        slack_message="message: {{ ds }}, {{ results_df }}",
    )
    # [END howto_operator_sql_to_slack_webhook]


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# NOTE! THIS FILE IS AUTOMATICALLY GENERATED AND WILL BE OVERWRITTEN!
#
# IF YOU WANT TO MODIFY THIS FILE, YOU SHOULD MODIFY THE TEMPLATE
# `get_provider_info_TEMPLATE.py.jinja2` IN the `dev/breeze/src/airflow_breeze/templates` DIRECTORY


def get_provider_info():
    return {
        "package-name": "apache-airflow-providers-slack",
        "name": "Slack",
        "description": "`Slack <https://slack.com/>`__ services integration including:\n\n  - `Slack API <https://api.slack.com/>`__\n  - `Slack Incoming Webhook <https://api.slack.com/messaging/webhooks>`__\n",
        "integrations": [
            {
                "integration-name": "Slack",
                "external-doc-url": "https://slack.com/",
                "logo": "/docs/integration-logos/Slack.png",
                "tags": ["service"],
            },
            {
                "integration-name": "Slack API",
                "external-doc-url": "https://api.slack.com/",
                "how-to-guide": ["/docs/apache-airflow-providers-slack/operators/slack_api.rst"],
                "tags": ["service"],
            },
            {
                "integration-name": "Slack Incoming Webhook",
                "external-doc-url": "https://api.slack.com/messaging/webhooks",
                "how-to-guide": ["/docs/apache-airflow-providers-slack/operators/slack_webhook.rst"],
                "tags": ["service"],
            },
        ],
        "operators": [
            {"integration-name": "Slack API", "python-modules": ["airflow.providers.slack.operators.slack"]},
            {
                "integration-name": "Slack Incoming Webhook",
                "python-modules": ["airflow.providers.slack.operators.slack_webhook"],
            },
        ],
        "hooks": [
            {"integration-name": "Slack API", "python-modules": ["airflow.providers.slack.hooks.slack"]},
            {
                "integration-name": "Slack Incoming Webhook",
                "python-modules": ["airflow.providers.slack.hooks.slack_webhook"],
            },
        ],
        "transfers": [
            {
                "source-integration-name": "Common SQL",
                "target-integration-name": "Slack",
                "python-module": "airflow.providers.slack.transfers.base_sql_to_slack",
            },
            {
                "source-integration-name": "Common SQL",
                "target-integration-name": "Slack API",
                "python-module": "airflow.providers.slack.transfers.sql_to_slack",
                "how-to-guide": "/docs/apache-airflow-providers-slack/operators/sql_to_slack.rst",
            },
            {
                "source-integration-name": "Common SQL",
                "target-integration-name": "Slack Incoming Webhook",
                "python-module": "airflow.providers.slack.transfers.sql_to_slack_webhook",
                "how-to-guide": "/docs/apache-airflow-providers-slack/operators/sql_to_slack_webhook.rst",
            },
        ],
        "connection-types": [
            {"hook-class-name": "airflow.providers.slack.hooks.slack.SlackHook", "connection-type": "slack"},
            {
                "hook-class-name": "airflow.providers.slack.hooks.slack_webhook.SlackWebhookHook",
                "connection-type": "slackwebhook",
            },
        ],
        "notifications": [
            "airflow.providers.slack.notifications.slack.SlackNotifier",
            "airflow.providers.slack.notifications.slack_webhook.SlackWebhookNotifier",
        ],
    }



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import annotations

import json
import os
from typing import Any
from unittest import mock
from unittest.mock import patch

import pytest
from slack_sdk.http_retry.builtin_handlers import ConnectionErrorRetryHandler, RateLimitErrorRetryHandler
from slack_sdk.webhook.async_client import AsyncWebhookClient
from slack_sdk.webhook.webhook_response import WebhookResponse

from airflow.models.connection import Connection
from airflow.providers.common.compat.sdk import AirflowException, AirflowNotFoundException
from airflow.providers.slack.hooks.slack_webhook import (
    SlackWebhookHook,
    async_check_webhook_response,
    check_webhook_response,
)

TEST_TOKEN = "T00000000/B00000000/XXXXXXXXXXXXXXXXXXXXXXXX"
TEST_WEBHOOK_URL = f"https://hooks.slack.com/services/{TEST_TOKEN}"
TEST_CUSTOM_SCHEMA = "http"
TEST_CUSTOM_ENDPOINT = "example.org/slack/webhooks"
TEST_CUSTOM_WEBHOOK_URL = f"{TEST_CUSTOM_SCHEMA}://{TEST_CUSTOM_ENDPOINT}/{TEST_TOKEN}"
TEST_CONN_ID = SlackWebhookHook.default_conn_name
CONN_TYPE = "slackwebhook"
TEST_CONN_ERROR_RETRY_HANDLER = ConnectionErrorRetryHandler(max_retry_count=42)
TEST_RATE_LIMIT_RETRY_HANDLER = RateLimitErrorRetryHandler()
MOCK_WEBHOOK_RESPONSE = WebhookResponse(url="foo://bar", status_code=200, body="ok", headers={})


@pytest.fixture(scope="module", autouse=True)
def slack_webhook_connections():
    """Create tests connections."""
    connections = [
        Connection(
            conn_id=SlackWebhookHook.default_conn_name,
            conn_type=CONN_TYPE,
            password=TEST_TOKEN,
        ),
        Connection(
            conn_id="conn_full_url_connection",
            conn_type=CONN_TYPE,
            password=TEST_WEBHOOK_URL,
        ),
        Connection(
            conn_id="conn_full_url_connection_with_host",
            conn_type=CONN_TYPE,
            host="http://example.org/hooks/",
            password=TEST_WEBHOOK_URL,
        ),
        Connection(
            conn_id="conn_host_with_schema",
            conn_type=CONN_TYPE,
            host="https://hooks.slack.com/services/",
            password=f"/{TEST_TOKEN}",
        ),
        Connection(
            conn_id="conn_host_without_schema",
            conn_type=CONN_TYPE,
            host="hooks.slack.com/services/",
            password=f"/{TEST_TOKEN}",
        ),
        Connection(
            conn_id="conn_parts",
            conn_type=CONN_TYPE,
            host="hooks.slack.com/services",
            schema="https",
            password=f"/{TEST_TOKEN}",
        ),
        Connection(
            conn_id="conn_deprecated_extra",
            conn_type=CONN_TYPE,
            host="https://hooks.slack.com/services/",
            extra={"webhook_token": TEST_TOKEN},
        ),
        Connection(
            conn_id="conn_custom_endpoint_1",
            conn_type=CONN_TYPE,
            schema=TEST_CUSTOM_SCHEMA,
            host=TEST_CUSTOM_ENDPOINT,
            password=TEST_TOKEN,
        ),
        Connection(
            conn_id="conn_custom_endpoint_2",
            conn_type=CONN_TYPE,
            host=f"{TEST_CUSTOM_SCHEMA}://{TEST_CUSTOM_ENDPOINT}",
            password=TEST_TOKEN,
        ),
        Connection(
            conn_id="conn_custom_endpoint_3",
            conn_type=CONN_TYPE,
            password=TEST_CUSTOM_WEBHOOK_URL,
        ),
        Connection(
            conn_id="conn_empty",
            conn_type=CONN_TYPE,
        ),
        Connection(
            conn_id="conn_password_empty_1",
            conn_type=CONN_TYPE,
            host="https://hooks.slack.com/services/",
        ),
        Connection(
            conn_id="conn_password_empty_2",
            conn_type=CONN_TYPE,
            schema="http",
            host="some.netloc",
        ),
        # Not supported anymore
        Connection(conn_id="conn_token_in_host_1", conn_type=CONN_TYPE, host=TEST_WEBHOOK_URL),
        Connection(
            conn_id="conn_token_in_host_2",
            conn_type=CONN_TYPE,
            schema="https",
            host="hooks.slack.com/services/T00000000/B00000000/XXXXXXXXXXXXXXXXXXXXXXXX",
        ),
    ]
    with pytest.MonkeyPatch.context() as mp:
        for conn in connections:
            mp.setenv(f"AIRFLOW_CONN_{conn.conn_id.upper()}", conn.get_uri())
        yield


class TestCheckWebhookResponseDecorator:
    def test_ok_response(self):
        """Test OK response."""

        @check_webhook_response
        def decorated():
            return MOCK_WEBHOOK_RESPONSE

        assert decorated() is MOCK_WEBHOOK_RESPONSE

    @pytest.mark.parametrize(
        ("status_code", "body"),
        [
            (400, "invalid_payload"),
            (403, "action_prohibited"),
            (404, "channel_not_found"),
            (410, "channel_is_archived"),
            (500, "rollup_error"),
            (418, "i_am_teapot"),
        ],
    )
    def test_error_response(self, status_code, body):
        """Test error response."""
        test_response = WebhookResponse(url="foo://bar", status_code=status_code, body=body, headers={})

        @check_webhook_response
        def decorated():
            return test_response

        error_message = rf"Response body: '{body}', Status Code: {status_code}\."
        with pytest.raises(AirflowException, match=error_message):
            assert decorated()


class TestAsyncCheckWebhookResponseDecorator:
    @pytest.mark.asyncio
    async def test_ok_response(self):
        """Test async decorator with OK response."""

        @async_check_webhook_response
        async def decorated():
            return MOCK_WEBHOOK_RESPONSE

        assert await decorated() is MOCK_WEBHOOK_RESPONSE

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ("status_code", "body"),
        [
            (400, "invalid_payload"),
            (403, "action_prohibited"),
            (404, "channel_not_found"),
            (410, "channel_is_archived"),
            (500, "rollup_error"),
            (418, "i_am_teapot"),
        ],
    )
    async def test_error_response(self, status_code, body):
        """Test async decorator with error response."""
        test_response = WebhookResponse(url="foo://bar", status_code=status_code, body=body, headers={})

        @async_check_webhook_response
        async def decorated():
            return test_response

        error_message = rf"Response body: '{body}', Status Code: {status_code}\."
        with pytest.raises(AirflowException, match=error_message):
            await decorated()


class TestSlackWebhookHook:
    @pytest.mark.parametrize(
        "conn_id",
        [
            TEST_CONN_ID,
            "conn_full_url_connection",
            "conn_full_url_connection_with_host",
            "conn_host_with_schema",
            "conn_host_without_schema",
            "conn_parts",
        ],
    )
    def test_construct_webhook_url(self, conn_id):
        """Test valid connections."""
        hook = SlackWebhookHook(slack_webhook_conn_id=conn_id)
        conn_params = hook._get_conn_params()
        assert not hasattr(hook, "http_conn_id")
        assert not hasattr(hook, "webhook_token")
        assert "url" in conn_params
        assert conn_params["url"] == TEST_WEBHOOK_URL

    def test_ignore_webhook_token(self):
        """Test that we only use token from Slack API Connection ID."""
        with pytest.warns(
            UserWarning, match="Provide `webhook_token` as part of .* parameters is disallowed"
        ):
            hook = SlackWebhookHook(slack_webhook_conn_id=TEST_CONN_ID, webhook_token="foo-bar")
        assert "webhook_token" not in hook.extra_client_args
        assert hook._get_conn_params()["url"] == TEST_WEBHOOK_URL

    @pytest.mark.parametrize("conn_id", ["conn_token_in_host_1", "conn_token_in_host_2"])
    def test_wrong_connections(self, conn_id):
        """Test previously valid connections, but now support of it is dropped."""
        hook = SlackWebhookHook(slack_webhook_conn_id=conn_id)
        with pytest.raises(AirflowNotFoundException, match="does not contain password"):
            hook._get_conn_params()

    @pytest.mark.parametrize(
        "conn_id", ["conn_custom_endpoint_1", "conn_custom_endpoint_2", "conn_custom_endpoint_3"]
    )
    def test_construct_webhook_url_with_non_default_host(self, conn_id):
        """Test valid connections with endpoint != https://hooks.slack.com/hooks."""
        hook = SlackWebhookHook(slack_webhook_conn_id=conn_id)
        conn_params = hook._get_conn_params()
        assert "url" in conn_params
        assert conn_params["url"] == TEST_CUSTOM_WEBHOOK_URL

    @pytest.mark.parametrize(
        "conn_id",
        [
            "conn_empty",
            "conn_password_empty_1",
            "conn_password_empty_2",
        ],
    )
    def test_no_password_in_connection_field(self, conn_id):
        """Test connection which missing password field in connection."""
        hook = SlackWebhookHook(slack_webhook_conn_id=conn_id)
        error_message = r"Connection ID .* does not contain password \(Slack Webhook Token\)"
        with pytest.raises(AirflowNotFoundException, match=error_message):
            hook._get_conn_params()

    @pytest.mark.parametrize(
        ("hook_config", "conn_extra", "expected"),
        [
            (  # Test Case: hook config
                {
                    "timeout": 42,
                    "proxy": "https://hook-proxy:1234",
                    "retry_handlers": [TEST_CONN_ERROR_RETRY_HANDLER, TEST_RATE_LIMIT_RETRY_HANDLER],
                },
                {},
                {
                    "timeout": 42,
                    "proxy": "https://hook-proxy:1234",
                    "retry_handlers": [TEST_CONN_ERROR_RETRY_HANDLER, TEST_RATE_LIMIT_RETRY_HANDLER],
                },
            ),
            (  # Test Case: connection config
                {},
                {
                    "timeout": 9000,
                    "proxy": "https://conn-proxy:4321",
                },
                {
                    "timeout": 9000,
                    "proxy": "https://conn-proxy:4321",
                },
            ),
            (  # Test Case: Connection from the UI
                {},
                {
                    "extra__slackwebhook__timeout": 9000,
                    "extra__slackwebhook__proxy": "https://conn-proxy:4321",
                },
                {
                    "timeout": 9000,
                    "proxy": "https://conn-proxy:4321",
                },
            ),
            (  # Test Case: Merge configs - hook args overwrite conn config
                {
                    "timeout": 1,
                    "proxy": "https://hook-proxy:777",
                },
                {
                    "timeout": 9000,
                    "proxy": "https://conn-proxy:4321",
                },
                {
                    "timeout": 1,
                    "proxy": "https://hook-proxy:777",
                },
            ),
            (  # Test Case: Merge configs - resolve config
                {
                    "timeout": 1,
                },
                {
                    "timeout": 9000,
                    "proxy": "https://conn-proxy:4334",
                },
                {
                    "timeout": 1,
                    "proxy": "https://conn-proxy:4334",
                },
            ),
            (  # Test Case: empty configs
                {},
                {},
                {},
            ),
            (  # Test Case: extra_client_args
                {"foo": "bar"},
                {},
                {"foo": "bar"},
            ),
            (  # Test Case: ignored not expected connection extra
                {},
                {"spam": "egg"},
                {},
            ),
        ],
    )
    @mock.patch("airflow.providers.slack.hooks.slack_webhook.WebhookClient")
    def test_client_configuration(
        self, mock_webhook_client_cls, hook_config, conn_extra, expected: dict[str, Any]
    ):
        """Test read/parse/merge WebhookClient config from connection and hook arguments."""
        expected["url"] = TEST_WEBHOOK_URL
        test_conn = Connection(
            conn_id="test-slack-incoming-webhook-conn",
            conn_type=CONN_TYPE,
            password=TEST_WEBHOOK_URL,
            extra=conn_extra,
        )
        test_conn_env = f"AIRFLOW_CONN_{test_conn.conn_id.upper()}"
        mock_webhook_client = mock_webhook_client_cls.return_value

        with mock.patch.dict("os.environ", values={test_conn_env: test_conn.get_uri()}):
            hook = SlackWebhookHook(slack_webhook_conn_id=test_conn.conn_id, **hook_config)
            expected["logger"] = hook.log
            conn_params = hook._get_conn_params()
            assert conn_params == expected

            client = hook.client
            assert client == mock_webhook_client
            assert hook.get_conn() == mock_webhook_client
            assert hook.get_conn() is client  # cached
            mock_webhook_client_cls.assert_called_once_with(**expected)

    @pytest.mark.parametrize("headers", [None, {"User-Agent": "Airflow"}])
    @pytest.mark.parametrize(
        "send_body",
        [
            {"text": "Test Text"},
            {"text": "Fallback Text", "blocks": ["Dummy Block"]},
            {"text": "Fallback Text", "blocks": ["Dummy Block"], "unfurl_media": True, "unfurl_links": True},
        ],
    )
    @mock.patch("airflow.providers.slack.hooks.slack_webhook.WebhookClient")
    def test_hook_send_dict(self, mock_webhook_client_cls, send_body, headers):
        """Test `SlackWebhookHook.send_dict` method."""
        mock_webhook_client = mock_webhook_client_cls.return_value
        mock_webhook_client_send_dict = mock_webhook_client.send_dict
        mock_webhook_client_send_dict.return_value = MOCK_WEBHOOK_RESPONSE

        hook = SlackWebhookHook(slack_webhook_conn_id=TEST_CONN_ID)
        # Test with regular dictionary
        hook.send_dict(body=send_body, headers=headers)
        mock_webhook_client_send_dict.assert_called_once_with(send_body, headers=headers)

        # Test with JSON-string
        mock_webhook_client_send_dict.reset_mock()
        hook.send_dict(body=json.dumps(send_body), headers=headers)
        mock_webhook_client_send_dict.assert_called_once_with(send_body, headers=headers)

    @pytest.mark.parametrize("send_body", [("text", "Test Text"), 42, "null", "42"])
    @mock.patch("airflow.providers.slack.hooks.slack_webhook.WebhookClient")
    def test_hook_send_dict_invalid_type(self, mock_webhook_client_cls, send_body):
        """Test invalid body type for `SlackWebhookHook.send_dict` method."""
        mock_webhook_client = mock_webhook_client_cls.return_value
        mock_webhook_client_send_dict = mock_webhook_client.send_dict
        mock_webhook_client_send_dict.return_value = MOCK_WEBHOOK_RESPONSE

        hook = SlackWebhookHook(slack_webhook_conn_id=TEST_CONN_ID)
        with pytest.raises(TypeError, match=r"Body expected dictionary, got .*\."):
            hook.send_dict(body=send_body)
        assert mock_webhook_client_send_dict.assert_not_called

    @pytest.mark.parametrize("json_string", ["{'text': 'Single quotes'}", '{"text": "Missing }"'])
    @mock.patch("airflow.providers.slack.hooks.slack_webhook.WebhookClient")
    def test_hook_send_dict_invalid_json_string(self, mock_webhook_client_cls, json_string):
        """Test invalid JSON-string passed to `SlackWebhookHook.send_dict` method."""
        mock_webhook_client = mock_webhook_client_cls.return_value
        mock_webhook_client_send_dict = mock_webhook_client.send_dict
        mock_webhook_client_send_dict.return_value = MOCK_WEBHOOK_RESPONSE

        hook = SlackWebhookHook(slack_webhook_conn_id=TEST_CONN_ID)
        error_message = r"Body expected valid JSON string, got .*\. Original error:.*"
        with pytest.raises(AirflowException, match=error_message):
            hook.send_dict(body=json_string)
        assert mock_webhook_client_send_dict.assert_not_called

    @pytest.mark.parametrize(
        "legacy_attr",
        [
            "channel",
            "username",
            "icon_emoji",
            "icon_url",
        ],
    )
    @mock.patch("airflow.providers.slack.hooks.slack_webhook.WebhookClient")
    def test_hook_send_dict_legacy_slack_integration(self, mock_webhook_client_cls, legacy_attr):
        """Test `SlackWebhookHook.send_dict` warn users about Legacy Slack Integrations."""
        mock_webhook_client = mock_webhook_client_cls.return_value
        mock_webhook_client_send_dict = mock_webhook_client.send_dict
        mock_webhook_client_send_dict.return_value = MOCK_WEBHOOK_RESPONSE

        legacy_slack_integration_body = {legacy_attr: "test-value"}
        hook = SlackWebhookHook(slack_webhook_conn_id=TEST_CONN_ID)
        warning_message = (
            r"You cannot override the default channel \(chosen by the user who installed your app\), "
            r"username, or icon when you're using Incoming Webhooks to post messages\. "
            r"Instead, these values will always inherit from the associated Slack app configuration\. "
            r"See: .*\. It is possible to change this values only in "
            r"Legacy Slack Integration Incoming Webhook: .*"
        )
        with pytest.warns(UserWarning, match=warning_message):
            hook.send_dict(body=legacy_slack_integration_body)
        mock_webhook_client_send_dict.assert_called_once_with(legacy_slack_integration_body, headers=None)

    @pytest.mark.parametrize("headers", [None, {"User-Agent": "Airflow"}])
    @pytest.mark.parametrize(
        "send_params",
        [
            {"text": "Test Text"},
            {"text": "Fallback Text", "blocks": ["Dummy Block"]},
            {"text": "Fallback Text", "blocks": ["Dummy Block"], "unfurl_media": True, "unfurl_links": True},
            {"legacy": "value"},
        ],
    )
    @mock.patch("airflow.providers.slack.hooks.slack_webhook.SlackWebhookHook.send_dict")
    def test_hook_send(self, mock_hook_send_dict, send_params, headers):
        """Test `SlackWebhookHook.send` method."""
        hook = SlackWebhookHook(slack_webhook_conn_id=TEST_CONN_ID)
        hook.send(**send_params, headers=headers)
        mock_hook_send_dict.assert_called_once_with(body=send_params, headers=headers)

    @pytest.mark.parametrize("headers", [None, {"User-Agent": "Airflow"}])
    @pytest.mark.parametrize("unfurl_links", [None, False, True])
    @pytest.mark.parametrize("unfurl_media", [None, False, True])
    @mock.patch("airflow.providers.slack.hooks.slack_webhook.SlackWebhookHook.send")
    def test_hook_send_text(self, mock_hook_send, headers, unfurl_links, unfurl_media):
        """Test `SlackWebhookHook.send_text` method."""
        hook = SlackWebhookHook(slack_webhook_conn_id=TEST_CONN_ID)
        hook.send_text("Test Text", headers=headers, unfurl_links=unfurl_links, unfurl_media=unfurl_media)
        mock_hook_send.assert_called_once_with(
            text="Test Text", headers=headers, unfurl_links=unfurl_links, unfurl_media=unfurl_media
        )

    @pytest.mark.parametrize(
        "uri",
        [
            pytest.param(
                "a://:abc@?extra__slackwebhook__timeout=123&extra__slackwebhook__proxy=proxy",
                id="prefix",
            ),
            pytest.param("a://:abc@?timeout=123&proxy=proxy", id="no-prefix"),
        ],
    )
    def test_backcompat_prefix_works(self, uri):
        with patch.dict(in_dict=os.environ, AIRFLOW_CONN_MY_CONN=uri):
            hook = SlackWebhookHook(slack_webhook_conn_id="my_conn")
            params = hook._get_conn_params()
            assert params["url"] == "https://hooks.slack.com/services/abc"
            assert params["timeout"] == 123
            assert params["proxy"] == "proxy"

    def test_backcompat_prefix_both_causes_warning(self):
        with patch.dict(
            in_dict=os.environ,
            AIRFLOW_CONN_MY_CONN="a://:abc@?extra__slackwebhook__timeout=111&timeout=222",
        ):
            hook = SlackWebhookHook(slack_webhook_conn_id="my_conn")
            with pytest.warns(Warning, match="Using value for `timeout`"):
                params = hook._get_conn_params()
            assert params["timeout"] == 222

    def test_empty_string_ignored_prefixed(self):
        with patch.dict(
            in_dict=os.environ,
            AIRFLOW_CONN_MY_CONN=json.dumps(
                {
                    "password": "hi",
                    "extra": {"extra__slackwebhook__proxy": ""},
                }
            ),
        ):
            hook = SlackWebhookHook(slack_webhook_conn_id="my_conn")
            params = hook._get_conn_params()
            assert "proxy" not in params

    def test_empty_string_ignored_non_prefixed(self):
        with patch.dict(
            in_dict=os.environ,
            AIRFLOW_CONN_MY_CONN=json.dumps({"password": "hi", "extra": {"proxy": ""}}),
        ):
            hook = SlackWebhookHook(slack_webhook_conn_id="my_conn")
            params = hook._get_conn_params()
            assert "proxy" not in params


class TestSlackWebhookHookAsync:
    @pytest.mark.asyncio
    @mock.patch("airflow.providers.slack.hooks.slack_webhook.SlackWebhookHook._async_get_conn_params")
    async def test_async_client(self, mock_async_get_conn_params):
        """Test async_client property creates AsyncWebhookClient with correct params."""
        mock_async_get_conn_params.return_value = {"url": TEST_WEBHOOK_URL}

        hook = SlackWebhookHook(slack_webhook_conn_id=TEST_CONN_ID)
        client = await hook.get_async_client()

        assert isinstance(client, AsyncWebhookClient)
        assert client.url == TEST_WEBHOOK_URL
        mock_async_get_conn_params.assert_called_once()

    @pytest.mark.asyncio
    @pytest.mark.parametrize("headers", [None, {"User-Agent": "Airflow"}])
    @pytest.mark.parametrize(
        "send_body",
        [
            {"text": "Test Text"},
            {"text": "Fallback Text", "blocks": ["Dummy Block"]},
            {"text": "Fallback Text", "blocks": ["Dummy Block"], "unfurl_media": True, "unfurl_links": True},
        ],
    )
    @mock.patch("airflow.providers.slack.hooks.slack_webhook.AsyncWebhookClient")
    @mock.patch("airflow.providers.slack.hooks.slack_webhook.SlackWebhookHook._async_get_conn_params")
    async def test_async_send_dict(
        self, mock_async_get_conn_params, mock_async_webhook_client_cls, send_body, headers
    ):
        """Test async_send_dict method with dict input."""
        mock_async_get_conn_params.return_value = {"url": TEST_WEBHOOK_URL}
        mock_async_client = mock_async_webhook_client_cls.return_value
        mock_async_client.send_dict = mock.AsyncMock(return_value=MOCK_WEBHOOK_RESPONSE)

        hook = SlackWebhookHook(slack_webhook_conn_id=TEST_CONN_ID)
        resp = await hook.async_send_dict(body=send_body, headers=headers)

        assert resp == MOCK_WEBHOOK_RESPONSE
        mock_async_client.send_dict.assert_called_once_with(send_body, headers=headers)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("headers", [None, {"User-Agent": "Airflow"}])
    @pytest.mark.parametrize(
        "send_body",
        [
            {"text": "Test Text"},
            {"text": "Fallback Text", "blocks": ["Dummy Block"]},
            {"text": "Fallback Text", "blocks": ["Dummy Block"], "unfurl_media": True, "unfurl_links": True},
        ],
    )
    @mock.patch("airflow.providers.slack.hooks.slack_webhook.AsyncWebhookClient")
    @mock.patch("airflow.providers.slack.hooks.slack_webhook.SlackWebhookHook._async_get_conn_params")
    async def test_async_send_dict_json_string(
        self, mock_async_get_conn_params, mock_async_webhook_client_cls, send_body, headers
    ):
        """Test async_send_dict method with JSON string input."""
        mock_async_get_conn_params.return_value = {"url": TEST_WEBHOOK_URL}
        mock_async_client = mock_async_webhook_client_cls.return_value
        mock_async_client.send_dict = mock.AsyncMock(return_value=MOCK_WEBHOOK_RESPONSE)

        hook = SlackWebhookHook(slack_webhook_conn_id=TEST_CONN_ID)
        resp = await hook.async_send_dict(body=json.dumps(send_body), headers=headers)

        assert resp == MOCK_WEBHOOK_RESPONSE
        mock_async_client.send_dict.assert_called_once_with(send_body, headers=headers)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("headers", [None, {"User-Agent": "Airflow"}])
    @pytest.mark.parametrize(
        "send_params",
        [
            {"text": "Test Text"},
            {"text": "Fallback Text", "blocks": ["Dummy Block"]},
            {"text": "Fallback Text", "blocks": ["Dummy Block"], "unfurl_media": True, "unfurl_links": True},
            {"legacy": "value"},
        ],
    )
    @mock.patch("airflow.providers.slack.hooks.slack_webhook.SlackWebhookHook.async_send_dict")
    async def test_async_send(self, mock_async_send_dict, send_params, headers):
        """Test at async_send method."""
        mock_async_send_dict.return_value = MOCK_WEBHOOK_RESPONSE

        hook = SlackWebhookHook(slack_webhook_conn_id=TEST_CONN_ID)
        resp = await hook.async_send(**send_params, headers=headers)

        assert resp == MOCK_WEBHOOK_RESPONSE
        mock_async_send_dict.assert_called_once_with(body=send_params, headers=headers)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
import warnings
from collections.abc import Sequence
from functools import cached_property
from typing import TYPE_CHECKING, Any, Literal

from airflow.exceptions import AirflowProviderDeprecationWarning
from airflow.providers.common.compat.sdk import BaseOperator
from airflow.providers.slack.hooks.slack import SlackHook

if TYPE_CHECKING:
    from slack_sdk.http_retry import RetryHandler

    from airflow.providers.common.compat.sdk import Context


class SlackAPIOperator(BaseOperator):
    """
    Base Slack Operator class.

    :param slack_conn_id: :ref:`Slack API Connection <howto/connection:slack>`
        which its password is Slack API token.
    :param method: The Slack API Method to Call (https://api.slack.com/methods).
    :param api_params: API Method call parameters (https://api.slack.com/methods). Optional
    :param timeout: The maximum number of seconds the client will wait to connect
        and receive a response from Slack. Optional
    :param base_url: A string representing the Slack API base URL. Optional
    :param proxy: Proxy to make the Slack API call. Optional
    :param retry_handlers: List of handlers to customize retry logic in ``slack_sdk.WebClient``. Optional
    """

    def __init__(
        self,
        *,
        slack_conn_id: str = SlackHook.default_conn_name,
        method: str,
        api_params: dict | None = None,
        base_url: str | None = None,
        proxy: str | None = None,
        timeout: int | None = None,
        retry_handlers: list[RetryHandler] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.slack_conn_id = slack_conn_id
        self.method = method
        self.api_params = api_params
        self.base_url = base_url
        self.timeout = timeout
        self.proxy = proxy
        self.retry_handlers = retry_handlers

    @cached_property
    def hook(self) -> SlackHook:
        """Slack Hook."""
        return SlackHook(
            slack_conn_id=self.slack_conn_id,
            base_url=self.base_url,
            timeout=self.timeout,
            proxy=self.proxy,
            retry_handlers=self.retry_handlers,
        )

    def construct_api_call_params(self) -> Any:
        """
        Construct API call parameters used by the execute function.

        Allow templating on the source fields of the ``api_call_params`` dict
        before construction.

        Child classes should override this. Each SlackAPIOperator child class is
        responsible for having function set ``self.api_call_params`` with a dict
        of API call parameters (https://api.slack.com/methods)
        """
        raise NotImplementedError(
            "SlackAPIOperator should not be used directly. Chose one of the subclasses instead"
        )

    def execute(self, context: Context):
        if not self.method:
            msg = f"Expected non empty `method` attribute in {type(self).__name__!r}, but got {self.method!r}"
            raise ValueError(msg)
        if not self.api_params:
            self.construct_api_call_params()
        self.hook.call(self.method, json=self.api_params)


class SlackAPIPostOperator(SlackAPIOperator):
    """
    Post messages to a Slack channel.

    .. code-block:: python

        slack = SlackAPIPostOperator(
            task_id="post_hello",
            dag=dag,
            text="hello there!",
            channel="#random",
        )

    :param channel: channel in which to post message on slack name (#general) or
        ID (C12318391). (templated)
    :param username: Username that airflow will be posting to Slack as. (templated)
    :param text: message to send to slack. (templated)
    :param icon_url: URL to icon used for this message
    :param blocks: A list of blocks to send with the message. (templated)
        See https://api.slack.com/reference/block-kit/blocks
    :param attachments: (legacy) A list of attachments to send with the message. (templated)
        See https://api.slack.com/docs/attachments
    """

    template_fields: Sequence[str] = ("username", "text", "attachments", "blocks", "channel")
    ui_color = "#FFBA40"

    def __init__(
        self,
        channel: str = "#general",
        username: str = "Airflow",
        text: str = (
            "No message has been set.\n"
            "Here is a cat video instead\n"
            "https://www.youtube.com/watch?v=J---aiyznGQ"
        ),
        icon_url: str = (
            "https://raw.githubusercontent.com/apache/airflow/main/airflow-core/src/airflow/ui/public/pin_100.png"
        ),
        blocks: list | None = None,
        attachments: list | None = None,
        **kwargs,
    ) -> None:
        super().__init__(method="chat.postMessage", **kwargs)
        self.channel = channel
        self.username = username
        self.text = text
        self.icon_url = icon_url
        self.attachments = attachments or []
        self.blocks = blocks or []

    def construct_api_call_params(self) -> Any:
        self.api_params = {
            "channel": self.channel,
            "username": self.username,
            "text": self.text,
            "icon_url": self.icon_url,
            "attachments": json.dumps(self.attachments),
            "blocks": json.dumps(self.blocks),
        }


class SlackAPIFileOperator(SlackAPIOperator):
    """
    Send a file to a Slack channel.

    .. code-block:: python

        # Send file with filename and filetype
        slack_operator_file = SlackAPIFileOperator(
            task_id="slack_file_upload_1",
            dag=dag,
            slack_conn_id="slack",
            channels="#general,#random",
            initial_comment="Hello World!",
            filename="/files/dags/test.txt",
            filetype="txt",
        )

        # Send file content
        slack_operator_file_content = SlackAPIFileOperator(
            task_id="slack_file_upload_2",
            dag=dag,
            slack_conn_id="slack",
            channels="#general",
            initial_comment="Hello World!",
            content="file content in txt",
        )

    :param channels: Comma-separated list of channel names or IDs where the file will be shared.
        If set this argument to None, then file will send to associated workspace. (templated)
    :param initial_comment: message to send to slack. (templated)
    :param filename: name of the file (templated)
    :param filetype: slack filetype. (templated) See: https://api.slack.com/types/file#file_types
    :param content: file content. (templated)
    :param title: title of file. (templated)
    :param snippet_type: Syntax type for the snippet being uploaded.(templated)
    :param method_version: The version of the method of Slack SDK Client to be used, either "v1" or "v2".
    """

    template_fields: Sequence[str] = (
        "channels",
        "initial_comment",
        "filename",
        "filetype",
        "content",
        "title",
        "snippet_type",
    )
    ui_color = "#44BEDF"

    def __init__(
        self,
        channels: str | Sequence[str] | None = None,
        initial_comment: str | None = None,
        filename: str | None = None,
        filetype: str | None = None,
        content: str | None = None,
        title: str | None = None,
        method_version: Literal["v1", "v2"] | None = None,
        snippet_type: str | None = None,
        **kwargs,
    ) -> None:
        super().__init__(method="files.upload", **kwargs)
        self.channels = channels
        self.initial_comment = initial_comment
        self.filename = filename
        self.filetype = filetype
        self.content = content
        self.title = title
        self.method_version = method_version
        self.snippet_type = snippet_type

        if self.filetype:
            warnings.warn(
                "The property `filetype` is no longer supported in slack_sdk and will be removed in a future release.",
                AirflowProviderDeprecationWarning,
                stacklevel=2,
            )

        if self.method_version:
            warnings.warn(
                "The property `method_version` is no longer required for `SlackAPIFileOperator`, as slack_sdk is using the files_upload_v2 method by default.",
                AirflowProviderDeprecationWarning,
                stacklevel=2,
            )

    def execute(self, context: Context):
        self.hook.send_file_v1_to_v2(
            channels=self.channels,
            # For historical reason SlackAPIFileOperator use filename as reference to file
            file=self.filename,
            content=self.content,
            initial_comment=self.initial_comment,
            title=self.title,
            snippet_type=self.snippet_type,
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import warnings
from collections.abc import Sequence
from typing import Any

try:
    from airflow.sdk.definitions._internal.types import NOTSET
except ImportError:
    from airflow.utils.types import NOTSET  # type: ignore[attr-defined,no-redef]


class ConnectionExtraConfig:
    """
    Helper class for rom Connection Extra.

    :param conn_type: Hook connection type.
    :param conn_id: Connection ID uses for appropriate error messages.
    :param extra: Connection extra dictionary.
    """

    def __init__(self, conn_type: str, conn_id: str | None = None, extra: dict[str, Any] | None = None):
        super().__init__()
        self.conn_type = conn_type
        self.conn_id = conn_id
        self.extra = extra or {}

    def get(self, field, default: Any = NOTSET):
        """
        Get specified field from Connection Extra.

        :param field: Connection extra field name.
        :param default: If specified then use as default value if field not present in Connection Extra.
        """
        backcompat_key = f"extra__{self.conn_type}__{field}"
        if self.extra.get(field) not in (None, ""):
            if self.extra.get(backcompat_key) not in (None, ""):
                warnings.warn(
                    f"Conflicting params `{field}` and `{backcompat_key}` found in extras for conn "
                    f"{self.conn_id}. Using value for `{field}`.  Please ensure this is the correct value "
                    f"and remove the backcompat key `{backcompat_key}`.",
                    UserWarning,
                    stacklevel=2,
                )
            return self.extra[field]
        if backcompat_key in self.extra and self.extra[backcompat_key] not in (None, ""):
            # Addition validation with non-empty required for connection which created in the UI
            # in Airflow 2.2. In these connections always present key-value pair for all prefixed extras
            # even if user do not fill this fields.
            # In additional fields from `wtforms.IntegerField` might contain None value.
            # E.g.: `{'extra__slackwebhook__proxy': '', 'extra__slackwebhook__timeout': None}`
            # From Airflow 2.3, using the prefix is no longer required.
            return self.extra[backcompat_key]
        if default is NOTSET:
            raise KeyError(
                f"Couldn't find {backcompat_key!r} or {field!r} "
                f"in Connection ({self.conn_id!r}) Extra and no default value specified."
            )
        return default

    def getint(self, field, default: Any = NOTSET) -> Any:
        """
        Get specified field from Connection Extra and evaluate as integer.

        :param field: Connection extra field name.
        :param default: If specified then use as default value if field not present in Connection Extra.
        """
        value = self.get(field=field, default=default)
        if value != default:
            value = int(value)
        return value


def parse_filename(
    filename: str, supported_file_formats: Sequence[str], fallback: str | None = None
) -> tuple[str, str | None]:
    """
    Parse filetype and compression from given filename.

    :param filename: filename to parse.
    :param supported_file_formats: list of supported file extensions.
    :param fallback: fallback to given file format.
    :returns: filetype and compression (if specified)
    """
    if not filename:
        raise ValueError("Expected 'filename' parameter is missing.")
    if fallback and fallback not in supported_file_formats:
        raise ValueError(f"Invalid fallback value {fallback!r}, expected one of {supported_file_formats}.")

    parts = filename.rsplit(".", 2)
    try:
        if len(parts) == 1:
            raise ValueError(f"No file extension specified in filename {filename!r}.")
        if parts[-1] in supported_file_formats:
            return parts[-1], None
        if len(parts) == 2:
            raise ValueError(
                f"Unsupported file format {parts[-1]!r}, expected one of {supported_file_formats}."
            )
        if parts[-2] not in supported_file_formats:
            raise ValueError(
                f"Unsupported file format '{parts[-2]}.{parts[-1]}', "
                f"expected one of {supported_file_formats} with compression extension."
            )
        return parts[-2], parts[-1]
    except ValueError as ex:
        if fallback:
            return fallback, None
        raise ex from None



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from unittest import mock

import pytest

from airflow.providers.slack.operators.slack import (
    SlackAPIFileOperator,
    SlackAPIOperator,
    SlackAPIPostOperator,
)

SLACK_API_TEST_CONNECTION_ID = "test_slack_conn_id"
DEFAULT_HOOKS_PARAMETERS = {"base_url": None, "timeout": None, "proxy": None, "retry_handlers": None}


class TestSlackAPIOperator:
    @mock.patch("airflow.providers.slack.operators.slack.SlackHook")
    @pytest.mark.parametrize(
        ("slack_op_kwargs", "hook_extra_kwargs"),
        [
            pytest.param({}, DEFAULT_HOOKS_PARAMETERS, id="default-hook-parameters"),
            pytest.param(
                {
                    "base_url": "https://foo.bar",
                    "timeout": 42,
                    "proxy": "http://spam.egg",
                    "retry_handlers": [],
                },
                {
                    "base_url": "https://foo.bar",
                    "timeout": 42,
                    "proxy": "http://spam.egg",
                    "retry_handlers": [],
                },
                id="with-extra-hook-parameters",
            ),
        ],
    )
    def test_hook(self, mock_slack_hook_cls, slack_op_kwargs, hook_extra_kwargs):
        mock_slack_hook = mock_slack_hook_cls.return_value
        op = SlackAPIOperator(
            task_id="test-mask-token",
            slack_conn_id=SLACK_API_TEST_CONNECTION_ID,
            method="foo.Bar",
            **slack_op_kwargs,
        )
        hook = op.hook
        assert hook == mock_slack_hook
        assert hook is op.hook
        mock_slack_hook_cls.assert_called_once_with(
            slack_conn_id=SLACK_API_TEST_CONNECTION_ID, **hook_extra_kwargs
        )


class TestSlackAPIPostOperator:
    def setup_method(self):
        self.test_username = "test_username"
        self.test_channel = "#test_slack_channel"
        self.test_text = "test_text"
        self.test_icon_url = "test_icon_url"
        self.test_attachments = [
            {
                "fallback": "Required plain-text summary of the attachment.",
                "color": "#36a64f",
                "pretext": "Optional text that appears above the attachment block",
                "author_name": "Bobby Tables",
                "author_link": "http://flickr.com/bobby/",
                "author_icon": "http://flickr.com/icons/bobby.jpg",
                "title": "Slack API Documentation",
                "title_link": "https://api.slack.com/",
                "text": "Optional text that appears within the attachment",
                "fields": [{"title": "Priority", "value": "High", "short": "false"}],
                "image_url": "http://my-website.com/path/to/image.jpg",
                "thumb_url": "http://example.com/path/to/thumb.png",
                "footer": "Slack API",
                "footer_icon": "https://platform.slack-edge.com/img/default_application_icon.png",
                "ts": 123456789,
            }
        ]
        self.test_blocks = [
            {
                "type": "section",
                "text": {
                    "text": "A message *with some bold text* and _some italicized text_.",
                    "type": "mrkdwn",
                },
                "fields": [
                    {"type": "mrkdwn", "text": "High"},
                    {"type": "plain_text", "emoji": True, "text": "String"},
                ],
            }
        ]
        self.test_attachments_in_json = json.dumps(self.test_attachments)
        self.test_blocks_in_json = json.dumps(self.test_blocks)
        self.test_api_params = {"key": "value"}

        self.expected_method = "chat.postMessage"
        self.expected_api_params = {
            "channel": self.test_channel,
            "username": self.test_username,
            "text": self.test_text,
            "icon_url": self.test_icon_url,
            "attachments": self.test_attachments_in_json,
            "blocks": self.test_blocks_in_json,
        }

    def __construct_operator(self, test_slack_conn_id, test_api_params=None):
        return SlackAPIPostOperator(
            task_id="slack",
            username=self.test_username,
            slack_conn_id=test_slack_conn_id,
            channel=self.test_channel,
            text=self.test_text,
            icon_url=self.test_icon_url,
            attachments=self.test_attachments,
            blocks=self.test_blocks,
            api_params=test_api_params,
        )

    def test_init_with_valid_params(self):
        slack_api_post_operator = self.__construct_operator(
            SLACK_API_TEST_CONNECTION_ID, self.test_api_params
        )
        assert slack_api_post_operator.slack_conn_id == SLACK_API_TEST_CONNECTION_ID
        assert slack_api_post_operator.method == self.expected_method
        assert slack_api_post_operator.text == self.test_text
        assert slack_api_post_operator.channel == self.test_channel
        assert slack_api_post_operator.api_params == self.test_api_params
        assert slack_api_post_operator.username == self.test_username
        assert slack_api_post_operator.icon_url == self.test_icon_url
        assert slack_api_post_operator.attachments == self.test_attachments
        assert slack_api_post_operator.blocks == self.test_blocks
        assert not hasattr(slack_api_post_operator, "token")

    @mock.patch("airflow.providers.slack.operators.slack.SlackHook")
    def test_api_call_params_with_default_args(self, mock_hook):
        slack_api_post_operator = SlackAPIPostOperator(
            task_id="slack",
            username=self.test_username,
            slack_conn_id=SLACK_API_TEST_CONNECTION_ID,
        )

        slack_api_post_operator.execute({})

        expected_api_params = {
            "channel": "#general",
            "username": self.test_username,
            "text": "No message has been set.\n"
            "Here is a cat video instead\n"
            "https://www.youtube.com/watch?v=J---aiyznGQ",
            "icon_url": "https://raw.githubusercontent.com/apache/"
            "airflow/main/airflow-core/src/airflow/ui/public/pin_100.png",
            "attachments": "[]",
            "blocks": "[]",
        }
        assert expected_api_params == slack_api_post_operator.api_params


class TestSlackAPIFileOperator:
    def setup_method(self):
        self.test_username = "test_username"
        self.test_channel = "#test_slack_channel"
        self.test_initial_comment = "test text file test_filename.txt"
        self.filename = "test_filename.txt"
        self.test_content = "This is a test text file!"
        self.test_api_params = {"key": "value"}
        self.expected_method = "files.upload"
        self.test_snippet_type = "text"

    def __construct_operator(self, test_slack_conn_id, test_api_params=None):
        return SlackAPIFileOperator(
            task_id="slack",
            slack_conn_id=test_slack_conn_id,
            channels=self.test_channel,
            initial_comment=self.test_initial_comment,
            filename=self.filename,
            content=self.test_content,
            api_params=test_api_params,
            snippet_type=self.test_snippet_type,
        )

    def test_init_with_valid_params(self):
        slack_api_post_operator = self.__construct_operator(
            SLACK_API_TEST_CONNECTION_ID, self.test_api_params
        )
        assert slack_api_post_operator.slack_conn_id == SLACK_API_TEST_CONNECTION_ID
        assert slack_api_post_operator.method == self.expected_method
        assert slack_api_post_operator.initial_comment == self.test_initial_comment
        assert slack_api_post_operator.channels == self.test_channel
        assert slack_api_post_operator.api_params == self.test_api_params
        assert slack_api_post_operator.filename == self.filename
        assert slack_api_post_operator.filetype is None
        assert slack_api_post_operator.content == self.test_content
        assert slack_api_post_operator.snippet_type == self.test_snippet_type
        assert not hasattr(slack_api_post_operator, "token")

    @pytest.mark.parametrize("initial_comment", [None, "foo-bar"])
    @pytest.mark.parametrize("title", [None, "Spam Egg"])
    @pytest.mark.parametrize("snippet_type", [None, "text"])
    def test_api_call_params_with_content_args(self, initial_comment, title, snippet_type):
        op = SlackAPIFileOperator(
            task_id="slack",
            slack_conn_id=SLACK_API_TEST_CONNECTION_ID,
            content="test-content",
            channels="#test-channel",
            initial_comment=initial_comment,
            title=title,
            snippet_type=snippet_type,
        )
        with mock.patch(
            "airflow.providers.slack.operators.slack.SlackHook.send_file_v1_to_v2"
        ) as mock_send_file:
            op.execute({})
            mock_send_file.assert_called_once_with(
                channels="#test-channel",
                content="test-content",
                file=None,
                initial_comment=initial_comment,
                title=title,
                snippet_type=snippet_type,
            )

    @pytest.mark.parametrize("initial_comment", [None, "foo-bar"])
    @pytest.mark.parametrize("title", [None, "Spam Egg"])
    @pytest.mark.parametrize("snippet_type", [None, "text"])
    def test_api_call_params_with_file_args(self, initial_comment, title, snippet_type):
        op = SlackAPIFileOperator(
            task_id="slack",
            slack_conn_id=SLACK_API_TEST_CONNECTION_ID,
            channels="C1234567890",
            filename="/dev/null",
            initial_comment=initial_comment,
            title=title,
            snippet_type=snippet_type,
        )
        with mock.patch(
            "airflow.providers.slack.operators.slack.SlackHook.send_file_v1_to_v2"
        ) as mock_send_file:
            op.execute({})
            mock_send_file.assert_called_once_with(
                channels="C1234567890",
                content=None,
                file="/dev/null",
                initial_comment=initial_comment,
                title=title,
                snippet_type=snippet_type,
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os
from datetime import datetime

from airflow.models.dag import DAG
from airflow.providers.slack.operators.slack import SlackAPIFileOperator, SlackAPIPostOperator

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID")
DAG_ID = "slack_api_example_dag"
SLACK_API_CONN_ID = os.environ.get("SLACK_API_CONN_ID", "slack_conn_id")
SLACK_CHANNEL = os.environ.get("SLACK_CHANNEL", "#general")
IMAGE_URL = (
    "https://raw.githubusercontent.com/apache/airflow/main/airflow-core/src/airflow/ui/public/pin_100.png"
)

with DAG(
    dag_id=DAG_ID,
    schedule=None,
    start_date=datetime(2021, 1, 1),
    default_args={"slack_conn_id": SLACK_API_CONN_ID, "initial_comment": "Hello World!"},
    max_active_runs=1,
    tags=["example"],
) as dag:
    # [START slack_api_post_operator_text_howto_guide]
    slack_operator_post_text = SlackAPIPostOperator(
        task_id="slack_post_text",
        channel=SLACK_CHANNEL,
        text=(
            "Apache Airflow® is an open-source platform for developing, "
            "scheduling, and monitoring batch-oriented workflows."
        ),
    )
    # [END slack_api_post_operator_text_howto_guide]

    # [START slack_api_post_operator_blocks_howto_guide]
    slack_operator_post_blocks = SlackAPIPostOperator(
        task_id="slack_post_blocks",
        channel=SLACK_CHANNEL,
        blocks=[
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": (
                        "*<https://github.com/apache/airflow|Apache Airflow®>* "
                        "is an open-source platform for developing, scheduling, "
                        "and monitoring batch-oriented workflows."
                    ),
                },
                "accessory": {"type": "image", "image_url": IMAGE_URL, "alt_text": "Pinwheel"},
            }
        ],
        text="Fallback message",
    )
    # [END slack_api_post_operator_blocks_howto_guide]

    # [START slack_api_file_operator_howto_guide]
    slack_operator_file = SlackAPIFileOperator(
        task_id="slack_file_upload_1",
        channels=SLACK_CHANNEL,
        filename="/files/dags/test.txt",
    )
    # [END slack_api_file_operator_howto_guide]

    # [START slack_api_file_operator_content_howto_guide]
    slack_operator_file_content = SlackAPIFileOperator(
        task_id="slack_file_upload_2",
        channels=SLACK_CHANNEL,
        content="file content in txt",
    )
    # [END slack_api_file_operator_content_howto_guide]

    (
        slack_operator_post_text
        >> slack_operator_post_blocks
        >> slack_operator_file
        >> slack_operator_file_content
    )


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest

from airflow.providers.slack.utils import ConnectionExtraConfig, parse_filename


class TestConnectionExtra:
    @pytest.mark.parametrize("conn_type", ["slack", "slack_incoming_webhook"])
    def test_get_extra_field(self, conn_type):
        """Test get arguments from connection extra: prefixed and not."""
        extra_config = ConnectionExtraConfig(
            conn_type=conn_type,
            conn_id="test-conn-id",
            extra={"arg1": "foo", f"extra__{conn_type}__arg2": "bar"},
        )
        assert extra_config.get("arg1") == "foo"
        assert extra_config.get("arg2") == "bar"

    def test_missing_extra_field(self):
        """Test missing field in extra."""
        extra_config = ConnectionExtraConfig(conn_type="slack", conn_id="test-conn-id", extra={})
        error_message = (
            r"Couldn't find 'extra__slack__arg_missing' or 'arg_missing' "
            r"in Connection \('test-conn-id'\) Extra and no default value specified\."
        )
        with pytest.raises(KeyError, match=error_message):
            extra_config.get("arg_missing")

    @pytest.mark.parametrize("value", [0, False, "", None], ids=lambda x: f"bool_false_{type(x).__name__}")
    def test_default_extra_field(self, value):
        """Test default value for missing field in extra."""
        extra_config = ConnectionExtraConfig(conn_type="slack", extra={})
        assert extra_config.get("arg_missing", default=value) == value

    @pytest.mark.parametrize("conn_type", ["slack", "slack_incoming_webhook"])
    def test_both_prefixed_and_not_in_extra_field(self, conn_type):
        """Test resolve field from extra when both specified prefixed and not for single field."""
        extra_config = ConnectionExtraConfig(
            conn_type=conn_type,
            conn_id="test-conn-id",
            extra={"arg1": "foo", f"extra__{conn_type}__arg1": "bar"},
        )
        assert extra_config.get("arg1") == "foo"

    @pytest.mark.parametrize("conn_type", ["slack", "slack_incoming_webhook"])
    @pytest.mark.parametrize("empty_value", [None, ""])
    def test_prefixed_extra_created_in_ui_connections(self, conn_type, empty_value):
        """Test that empty strings or None values in UI ignored."""
        extra_config = ConnectionExtraConfig(
            conn_type=conn_type,
            conn_id="test-conn-id",
            extra={
                f"extra__{conn_type}__arg_missing": empty_value,
                "arg_extra": "bar",
                f"extra__{conn_type}__arg_extra": empty_value,
            },
        )
        error_message = (
            r"Couldn't find '.*' or '.*' in Connection \('.*'\) Extra and no default value specified\."
        )
        with pytest.raises(KeyError, match=error_message):
            # No fallback should raise an error
            extra_config.get("arg_missing")

        assert extra_config.get("arg_missing", default="foo") == "foo"
        assert extra_config.get("arg_extra") == "bar"

    def test_get_parse_int(self):
        extra_config = ConnectionExtraConfig(
            conn_type="slack",
            extra={
                "int_arg_1": "42",
                "int_arg_2": 9000,
            },
        )
        assert extra_config.getint("int_arg_1") == 42
        assert extra_config.getint("int_arg_2") == 9000


class TestParseFilename:
    SUPPORTED_FORMAT = ("so", "dll", "exe", "sh")

    def test_error_parse_without_extension(self):
        with pytest.raises(ValueError, match="No file extension specified in filename"):
            assert parse_filename("Untitled File", self.SUPPORTED_FORMAT)

    @pytest.mark.parametrize(
        ("filename", "expected_format"),
        [
            ("libc.so", "so"),
            ("kernel32.dll", "dll"),
            ("xxx.mp4.exe", "exe"),
            ("init.sh", "sh"),
        ],
    )
    def test_parse_first_level(self, filename, expected_format):
        assert parse_filename(filename, self.SUPPORTED_FORMAT) == (expected_format, None)

    @pytest.mark.parametrize("filename", ["New File.txt", "cats-memes.mp4"])
    def test_error_parse_first_level(self, filename):
        with pytest.raises(ValueError, match="Unsupported file format"):
            assert parse_filename(filename, self.SUPPORTED_FORMAT)

    @pytest.mark.parametrize(
        ("filename", "expected"),
        [
            ("libc.so.6", ("so", "6")),
            ("kernel32.dll.zip", ("dll", "zip")),
            ("explorer.exe.7z", ("exe", "7z")),
            ("init.sh.gz", ("sh", "gz")),
        ],
    )
    def test_parse_second_level(self, filename, expected):
        assert parse_filename(filename, self.SUPPORTED_FORMAT) == expected

    @pytest.mark.parametrize("filename", ["example.so.tar.gz", "w.i.e.r.d"])
    def test_error_parse_second_level(self, filename):
        with pytest.raises(ValueError, match="Unsupported file format.*with compression extension."):
            assert parse_filename(filename, self.SUPPORTED_FORMAT)

    @pytest.mark.parametrize("filename", ["Untitled File", "New File.txt", "example.so.tar.gz"])
    @pytest.mark.parametrize("fallback", SUPPORTED_FORMAT)
    def test_fallback(self, filename, fallback):
        assert parse_filename(filename, self.SUPPORTED_FORMAT, fallback) == (fallback, None)

    @pytest.mark.parametrize("filename", ["Untitled File", "New File.txt", "example.so.tar.gz"])
    def test_wrong_fallback(self, filename):
        with pytest.raises(ValueError, match="Invalid fallback value"):
            assert parse_filename(filename, self.SUPPORTED_FORMAT, "mp4")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Iterable, Mapping, Sequence
from typing import TYPE_CHECKING, Any

from tabulate import tabulate

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.slack.hooks.slack_webhook import SlackWebhookHook
from airflow.providers.slack.transfers.base_sql_to_slack import BaseSqlToSlackOperator

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


class SqlToSlackWebhookOperator(BaseSqlToSlackOperator):
    """
    Executes an SQL statement in a given SQL connection and sends the results to Slack Incoming Webhook.

    The results of the query are rendered into the 'slack_message' parameter as a Pandas
    dataframe using a JINJA variable called '{{ results_df }}'. The 'results_df' variable
    name can be changed by specifying a different 'results_df_name' parameter. The Tabulate
    library is added to the JINJA environment as a filter to allow the dataframe to be
    rendered nicely. For example, set 'slack_message' to {{ results_df |
    tabulate(tablefmt="pretty", headers="keys") }} to send the results to Slack as an ascii
    rendered table.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:SqlToSlackWebhookOperator`

    .. note::
        You cannot override the default channel (chosen by the user who installed your app),
        Instead, these values will always inherit from the associated Slack App configuration
        (`link <https://api.slack.com/messaging/webhooks#advanced_message_formatting>`_).
        It is possible to change this values only in `Legacy Slack Integration Incoming Webhook
        <https://api.slack.com/legacy/custom-integrations/messaging/webhooks#legacy-customizations>`_.

    .. warning::
        This hook intend to use `Slack Incoming Webhook` connection
        and might not work correctly with `Slack API` connection.

    :param sql: The SQL query to be executed (templated)
    :param slack_message: The templated Slack message to send with the data returned from the SQL connection.
        You can use the default JINJA variable {{ results_df }} to access the pandas dataframe containing the
        SQL results
    :param sql_conn_id: reference to a specific database.
    :param sql_hook_params: Extra config params to be passed to the underlying hook.
           Should match the desired hook constructor params.
    :param slack_webhook_conn_id: :ref:`Slack Incoming Webhook <howto/connection:slack>`
        connection id that has Incoming Webhook token in the password field.
    :param slack_channel: The channel to send message.
    :param results_df_name: The name of the JINJA template's dataframe variable, default is 'results_df'
    :param parameters: The parameters to pass to the SQL query
    """

    template_fields: Sequence[str] = ("sql", "slack_message")
    template_ext: Sequence[str] = (".sql", ".jinja", ".j2")
    template_fields_renderers = {"sql": "sql", "slack_message": "jinja"}
    times_rendered = 0

    def __init__(
        self,
        *,
        sql: str,
        sql_conn_id: str,
        slack_webhook_conn_id: str | None = None,
        sql_hook_params: dict | None = None,
        slack_channel: str | None = None,
        slack_message: str,
        results_df_name: str = "results_df",
        parameters: list | tuple | Mapping[str, Any] | None = None,
        **kwargs,
    ) -> None:
        if not slack_webhook_conn_id:
            raise ValueError("Got an empty `slack_webhook_conn_id` value.")
        super().__init__(
            sql=sql, sql_conn_id=sql_conn_id, sql_hook_params=sql_hook_params, parameters=parameters, **kwargs
        )

        self.slack_webhook_conn_id = slack_webhook_conn_id
        self.slack_channel = slack_channel
        self.slack_message = slack_message
        self.results_df_name = results_df_name
        self.kwargs = kwargs

    def _render_and_send_slack_message(self, context, df) -> None:
        # Put the dataframe into the context and render the JINJA template fields
        context[self.results_df_name] = df
        self.render_template_fields(context)

        slack_hook = self._get_slack_hook()
        self.log.info("Sending slack message: %s", self.slack_message)
        slack_hook.send(text=self.slack_message, channel=self.slack_channel)

    def _get_slack_hook(self) -> SlackWebhookHook:
        return SlackWebhookHook(
            slack_webhook_conn_id=self.slack_webhook_conn_id,
            proxy=self.slack_proxy,
            timeout=self.slack_timeout,
            retry_handlers=self.slack_retry_handlers,
        )

    def render_template_fields(self, context, jinja_env=None) -> None:
        # If this is the first render of the template fields, exclude slack_message from rendering since
        # the SQL results haven't been retrieved yet.
        if self.times_rendered == 0:
            fields_to_render: Iterable[str] = (x for x in self.template_fields if x != "slack_message")
        else:
            fields_to_render = self.template_fields

        if not jinja_env:
            jinja_env = self.get_template_env()

        # Add the tabulate library into the JINJA environment
        jinja_env.filters["tabulate"] = tabulate

        self._do_render_template_fields(self, fields_to_render, context, jinja_env, set())
        self.times_rendered += 1

    def execute(self, context: Context) -> None:
        if not isinstance(self.sql, str):
            raise AirflowException("Expected 'sql' parameter should be a string.")
        if self.sql is None or self.sql.strip() == "":
            raise AirflowException("Expected 'sql' parameter is missing.")
        if self.slack_message is None or self.slack_message.strip() == "":
            raise AirflowException("Expected 'slack_message' parameter is missing.")

        df = self._get_query_results()
        self._render_and_send_slack_message(context, df)

        self.log.debug("Finished sending SQL data to Slack")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import json
from collections.abc import Sequence
from functools import cached_property
from typing import TYPE_CHECKING

from airflow.providers.common.compat.notifier import BaseNotifier
from airflow.providers.slack.hooks.slack import SlackHook
from airflow.providers.slack.version_compat import AIRFLOW_V_3_1_PLUS

if TYPE_CHECKING:
    from slack_sdk.http_retry import RetryHandler

ICON_URL: str = (
    "https://raw.githubusercontent.com/apache/airflow/main/airflow-core/src/airflow/ui/public/pin_100.png"
)


class SlackNotifier(BaseNotifier):
    """
    Slack BaseNotifier.

    :param slack_conn_id: Slack API token (https://api.slack.com/web).
    :param text: The content of the message
    :param channel: The channel to send the message to. Optional
    :param username: The username to send the message as. Optional
    :param icon_url: The icon to use for the message. Optional
    :param blocks: A list of blocks to send with the message. Optional
    :param timeout: The maximum number of seconds the client will wait to connect
        and receive a response from Slack. Optional
    :param base_url: A string representing the Slack API base URL. Optional
    :param proxy: Proxy to make the Slack API call. Optional
    :param retry_handlers: List of handlers to customize retry logic in ``slack_sdk.WebClient``. Optional
    :param attachments: (legacy) A list of attachments to send with the message. Optional
    :param unfurl_links: Option to indicate whether text url should unfurl. Optional
    :param unfurl_media: Option to indicate whether media url should unfurl. Optional
    """

    template_fields = ("text", "channel", "username", "attachments", "blocks")

    def __init__(
        self,
        *,
        slack_conn_id: str = SlackHook.default_conn_name,
        text: str = "This is a default message",
        channel: str = "#general",
        username: str = "Airflow",
        icon_url: str = ICON_URL,
        attachments: Sequence = (),
        blocks: Sequence = (),
        base_url: str | None = None,
        proxy: str | None = None,
        timeout: int | None = None,
        retry_handlers: list[RetryHandler] | None = None,
        unfurl_links: bool = True,
        unfurl_media: bool = True,
        **kwargs,
    ):
        if AIRFLOW_V_3_1_PLUS:
            #  Support for passing context was added in 3.1.0
            super().__init__(**kwargs)
        else:
            super().__init__()
        self.slack_conn_id = slack_conn_id
        self.text = text
        self.channel = channel
        self.username = username
        self.icon_url = icon_url
        self.attachments = attachments
        self.blocks = blocks
        self.base_url = base_url
        self.timeout = timeout
        self.proxy = proxy
        self.retry_handlers = retry_handlers
        self.unfurl_links = unfurl_links
        self.unfurl_media = unfurl_media

    @cached_property
    def hook(self) -> SlackHook:
        """Slack Hook."""
        return SlackHook(
            slack_conn_id=self.slack_conn_id,
            base_url=self.base_url,
            timeout=self.timeout,
            proxy=self.proxy,
            retry_handlers=self.retry_handlers,
        )

    def notify(self, context):
        """Send a message to a Slack Channel."""
        api_call_params = {
            "channel": self.channel,
            "username": self.username,
            "text": self.text,
            "icon_url": self.icon_url,
            "attachments": json.dumps(self.attachments),
            "blocks": json.dumps(self.blocks),
            "unfurl_links": self.unfurl_links,
            "unfurl_media": self.unfurl_media,
        }
        self.hook.call("chat.postMessage", json=api_call_params)

    async def async_notify(self, context):
        """Send a message to a Slack Channel (async)."""
        api_call_params = {
            "channel": self.channel,
            "username": self.username,
            "text": self.text,
            "icon_url": self.icon_url,
            "attachments": json.dumps(self.attachments),
            "blocks": json.dumps(self.blocks),
            "unfurl_links": self.unfurl_links,
            "unfurl_media": self.unfurl_media,
        }
        await self.hook.async_call("chat.postMessage", json=api_call_params)


send_slack_notification = SlackNotifier
